{ Headers
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _header;

interface

uses
 _common;

const
 hdrNONE           = 0;
 hdrVALUE          = 1;
 hdrTEXT           = 2;
 hdrRELOCTBL       = 3;
 hdrFILLTO         = 4;
 hdrROUND          = 5;
 hdrMEMO           = 6;
 hdrMEMOWRITE      = 7;

 hdrBLOCKCNT       = 10;
 hdrLASTBLOCKSIZE  = 11;
 hdrRELOCCNT       = 12;
 hdrSIZE           = 13;
 hdrSTACK          = 14;

 hdrSEGOFFSET      = 32;
 hdrSEGSIZE        = 33;
 hdrEND            = 254;

type
 THeader = record
    Name           : string;
    IdentName      : string;
    HeaderSize     : TAsmValue;
    FillPatternOfs : TDefValue;
    FillPatternLen : TAsmValue;
 end;
 THeaderItem = record
    Header         : TDefValue;
    ID             : TDefValue;
    Type1          : byte;
    Type2          : byte;
    Value1         : TAsmValue;
    Value2         : TAsmValue;
    Opr            : byte;
    StrValue       : string;
 end;

var
 HeaderList        : array of THeader;
 HeaderCount       : TDefValue;

 HeaderItemList    : array of THeaderItem;
 HeaderItemCount   : TDefValue;
 NewHeaderItemList : array of THeaderItem;
 NewHeaderItemCount: TDefValue;

 RelocList         : array of TAsmValue;
 RelocCount        : TDefValue;

procedure InitHeaders;

function GetHeaderIndex(hNam: string): TDefValue;

procedure AddHeader(hdrName, hdrIdent: string;
 fiOfs: TDefValue; fiLen: TAsmValue);
procedure AddHeaderItem(hName: string; hdID: TDefValue;
 vaTyp1, vaTyp2: byte; val1, val2: TAsmValue;
 valOpr: byte; stringVal: string);
procedure AddNewHeaderItem(hdID: TDefValue;
 vaTyp1, vaTyp2: byte; val1, val2: TAsmValue;
 valOpr: byte; stringVal: string);
procedure AddReloc(ofs: TAsmValue);

procedure AddHeaderValue(hdrName: string; vTyp: byte; val: TAsmValue);
procedure AddHeaderText(hdrName, str: string);
procedure AddHeaderRelocTable(hdrName: string; vTyp: byte);
procedure AddHeaderFill(hdrName: string; cnt: TAsmValue);
procedure AddHeaderRound(hdrName: string; cnt: TAsmValue);
procedure AddHeaderMemo(hdrName: string; memLoc: TAsmValue);
procedure AddHeaderMemoWrite(hdrName: string; vTyp, mem1, mOpr, mem2: byte);
procedure AddHeaderWriteItem(hdrName: string; vTyp, item, opr: byte; val: TAsmValue);
procedure AddHeaderSegOffset(hdrName: string; vTyp: byte; segNam: string;
 opr: byte; val: TAsmValue);
procedure AddHeaderSegSize(hdrName: string; vTyp: byte; segNam: string;
 opr: byte; val: TAsmValue);

procedure PrepareHeader(indx: TDefValue);
procedure BuildHeader(indx: TDefValue);

implementation

uses
 SysUtils, _error, _emit, _func, _scanner, _parser,
 _mem, _conv, _ident, _seg, _output;

procedure InitHeaders;
begin
   SetLength(HeaderList, 1);
   HeaderCount:= 0;
   SetLength(HeaderItemList, 1);
   HeaderItemCount:= 0;
   SetLength(NewHeaderItemList, 1);
   NewHeaderItemCount:= 0;
   SetLength(RelocList, 1);
   RelocCount:= 0;
end;



function GetHeaderIndex(hNam: string): TDefValue;
var
 hIndx, i          : TDefValue;
begin
   hIndx:= NotFound;
   if HeaderCount > 0 then begin
      i:= 0;
      while i < HeaderCount do begin
         if HeaderList[i].IdentName = hNam then begin
            hIndx:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetHeaderIndex:= hIndx;
end;



procedure AddHeader(hdrName, hdrIdent: string;
 fiOfs: TDefValue; fiLen: TAsmValue);
var
 hFound            : boolean;
 i                 : TDefValue;
begin
   hFound:= false;
   if HeaderCount > 0 then begin
      i:= 0;
      while i < HeaderCount do begin
         if HeaderList[i].IdentName = hdrIdent then begin
	          hFound:= true;
            break;
         end;
         Inc(i);
      end;
   end;
   if not hFound then begin
      SetLength(HeaderList, HeaderCount + 1);
      HeaderList[HeaderCount].Name:= hdrName;
      HeaderList[HeaderCount].IdentName:= hdrIdent;
      HeaderList[HeaderCount].FillPatternOfs:= fiOfs;
      HeaderList[HeaderCount].FillPatternLen:= fiLen;
      HeaderList[HeaderCount].HeaderSize:= 0;
      Inc(HeaderCount);
   end else begin
      ErrorMsg('Header already exists [' + hdrIdent + ']!');
   end;
end;

procedure AddHeaderItem(hName: string; hdID: TDefValue;
 vaTyp1, vaTyp2: byte; val1, val2: TAsmValue;
 valOpr: byte; stringVal: string);
var
 indx              : TDefValue;
begin
   indx:= GetHeaderIndex(hName);
   if indx <> NotFound then begin
      SetLength(HeaderItemList, HeaderItemCount + 1);
      HeaderItemList[HeaderItemCount].Header:= indx;
      HeaderItemList[HeaderItemCount].ID:= hdID;
      HeaderItemList[HeaderItemCount].Type1:= vaTyp1;
      HeaderItemList[HeaderItemCount].Type2:= vaTyp2;
      HeaderItemList[HeaderItemCount].Value1:= val1;
      HeaderItemList[HeaderItemCount].Value2:= val2;
      HeaderItemList[HeaderItemCount].Opr:= valOpr;
      HeaderItemList[HeaderItemCount].StrValue:= stringVal;
      Inc(HeaderItemCount);
   end else begin
      ErrorMsg('Header not found [' + hName + ']!');
   end;
end;

procedure AddNewHeaderItem(hdID: TDefValue;
 vaTyp1, vaTyp2: byte; val1, val2: TAsmValue;
 valOpr: byte; stringVal: string);
begin
   SetLength(NewHeaderItemList, NewHeaderItemCount + 1);
   NewHeaderItemList[NewHeaderItemCount].Header:= NotFound;
   NewHeaderItemList[NewHeaderItemCount].ID:= hdID;
   NewHeaderItemList[NewHeaderItemCount].Type1:= vaTyp1;
   NewHeaderItemList[NewHeaderItemCount].Type2:= vaTyp2;
   NewHeaderItemList[NewHeaderItemCount].Value1:= val1;
   NewHeaderItemList[NewHeaderItemCount].Value2:= val2;
   NewHeaderItemList[NewHeaderItemCount].Opr:= valOpr;
   NewHeaderItemList[NewHeaderItemCount].StrValue:= stringVal;
   Inc(NewHeaderItemCount);
end;

procedure AddReloc(ofs: TAsmValue);
begin
   SetLength(RelocList, RelocCount + 1);
   RelocList[Reloccount]:= ofs;
   Inc(RelocCount);
end;



procedure AddHeaderValue(hdrName: string; vTyp: byte; val: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrVALUE, vTyp, tNONE, val, 0, oprNONE, '');
end;

procedure AddHeaderText(hdrName, str: string);
begin
   AddHeaderItem(hdrName, hdrTEXT, tNONE, tNONE, 0, 0, oprNONE, str);
end;

procedure AddHeaderRelocTable(hdrName: string; vTyp: byte);
begin
   AddHeaderItem(hdrName, hdrRELOCTBL, vTyp, tNONE, 0, 0, oprNONE, '');
end;

procedure AddHeaderFill(hdrName: string; cnt: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrFILLTO, tNONE, tNONE, cnt, 0, oprNONE, '');
end;

procedure AddHeaderRound(hdrName: string; cnt: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrROUND, tNONE, tNONE, cnt, 0, oprNONE, '');
end;

procedure AddHeaderMemo(hdrName: string; memLoc: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrMEMO, tNONE, tNONE, memLoc, 0, oprNONE, '');
end;

procedure AddHeaderMemoWrite(hdrName: string; vTyp, mem1, mOpr, mem2: byte);
begin
   AddHeaderItem(hdrName, hdrMEMOWRITE, vTyp, tNONE, mem1, mem2, mOpr, '');
end;

procedure AddHeaderWriteItem(hdrName: string; vTyp, item, opr: byte; val: TAsmValue);
begin
   AddHeaderItem(hdrName, item, vTyp, tNONE, val, 0, opr, '');
end;

procedure AddHeaderSegOffset(hdrName: string; vTyp: byte; segNam: string;
 opr: byte; val: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrSEGOFFSET, vTyp, tNONE, val, 0, opr, segNam);
end;

procedure AddHeaderSegSize(hdrName: string; vTyp: byte; segNam: string;
 opr: byte; val: TAsmValue);
begin
   AddHeaderItem(hdrName, hdrSEGSIZE, vTyp, tNONE, val, 0, opr, segNam);
end;



procedure PrepareHeader(indx: TDefValue);
var
 eHdr, iFnd        : boolean;
 i, c, r, p        : TDefValue;
 hdr, id           : TDefValue;
 vt1, vt2, opr, b  : byte;
 val1, val2, v     : TAsmValue;
 str               : string;
 siz               : integer;
 curSize           : TAsmValue;
 ptnOfs            : TDefValue;
 ptnLen, ptnPos    : TAsmValue;
 count             : TAsmValue;
begin
   eHdr:= false;
   if HeaderItemCount > 0 then begin
      EmitLn('------ START PREPARING HEADER [' + MakeTime(true) + '] ------');
      i:= 0;
      while i < HeaderItemCount do begin
         IsErr:= false;
         if eHdr then break;
         hdr:= HeaderItemList[i].Header;
         if hdr = indx then begin
            id:= HeaderItemList[i].ID;
            vt1:= HeaderItemList[i].Type1;
            vt2:= HeaderItemList[i].Type2;
            val1:= HeaderItemList[i].Value1;
            val2:= HeaderItemList[i].Value2;
            opr:= HeaderItemList[i].Opr;
            str:= HeaderItemList[i].StrValue;
            iFnd:= false;
	          if (id = hdrVALUE) and (not iFnd) then begin
               iFnd:= true;
               AddNewHeaderItem(hdrVALUE, vt1, vt2, val1, val2, opr, str);
               siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               HeaderList[indx].HeaderSize:= HeaderList[indx].HeaderSize + siz;
            end;
            if (id = hdrTEXT) and (not iFnd) then begin
	             iFnd:= true;
               siz:= Length(str);
               if siz > 0 then begin
                  c:= 1;
                  while c <= siz do begin
                     b:= ord(str[c]);
                     AddNewHeaderItem(hdrVALUE, tBYTE, tNONE, b, 0, tNONE, '');
                     inc(c);
                  end;
                  HeaderList[indx].HeaderSize:= HeaderList[indx].HeaderSize + siz;
               end;
            end;
            if (id = hdrRELOCTBL) and (not iFnd) then begin
               iFnd:= true;
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               if RelocCount > 0 then begin
                  r:= 0;
                  while r < RelocCount do begin
                     v:= RelocList[r];
                     AddNewHeaderItem(hdrVALUE, vt1, tNONE, v, 0, oprNONE, '');
                     HeaderList[indx].HeaderSize:= HeaderList[indx].HeaderSize + siz;
                     Inc(r);
                  end;
               end;
            end;
	          if (id = hdrFILLTO) and (not iFnd) then begin
               iFnd:= true;
               curSize:= HeaderList[indx].HeaderSize;
               ptnOfs:= HeaderList[indx].FillPatternOfs;
               ptnLen:= HeaderList[indx].FillPatternLen;
               ptnPos:= 0;
               p:= curSize+1;
               while p <= val1 do begin
                  b:= 0;
                  if ptnLen > 0 then begin
                     b:= MemData[ptnOfs+ptnPos];
                     Inc(ptnPos);
                     if ptnPos > (ptnLen-1) then ptnPos:= 0;
                  end;
                  AddNewHeaderItem(hdrVALUE, tBYTE, tNONE, b, 0, oprNONE, '');
                  Inc(HeaderList[indx].HeaderSize);
                  Inc(p);
               end;
            end;
            if (id = hdrROUND) and (not iFnd) then begin
               iFnd:= true;
               curSize:= HeaderList[indx].HeaderSize;
               count:= curSize div val1;
               count:= curSize - (count * val1);
               if (count > 0) and (count < val1) then begin
                  count:= val1 - count;
                  ptnOfs:= HeaderList[indx].FillPatternOfs;
                  ptnLen:= HeaderList[indx].FillPatternLen;
                  ptnPos:= 0;
                  p:= 0;
                  while p < count do begin
		                 b:= 0;
                     if ptnLen > 0 then begin
                        b:= MemData[ptnOfs+ptnPos];
                        Inc(ptnPos);
                        if ptnPos > (ptnLen-1) then ptnPos:= 0;
                     end;
                     AddNewHeaderItem(hdrVALUE, tBYTE, tNONE, b, 0, oprNONE, '');
                     Inc(HeaderList[indx].HeaderSize);
                     Inc(p);
                  end;
               end;
            end;
            if (id = hdrMEMO) and (not iFnd) then begin
               iFnd:= true;
               hdrMem[val1]:= HeaderList[indx].HeaderSize;
            end;
            if (id = hdrMEMOWRITE) and (not iFnd) then begin
               iFnd:= true;
               AddNewHeaderItem(hdrMEMOWRITE, vt1, tNONE, val1, val2, opr, '');
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
	             HeaderList[indx].HeaderSize:= HeaderList[indx].HeaderSize + siz;
            end;
            if (id = hdrEND) and (not iFnd) then begin
               iFnd:= true;
               eHdr:= true;
            end;
            if not iFnd then begin
               AddNewHeaderItem(id, vt1, vt2, val1, val2, opr, str);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               HeaderList[indx].HeaderSize:= HeaderList[indx].HeaderSize + siz;
            end;
         end;
         Inc(i);
      end;
      EmitLn('Header Size: ' + IntToStr(HeaderList[indx].HeaderSize));
      EmitLn('------ END PREPARING HEADER [' + MakeTime(true) + '] ------');
   end;
end;

procedure BuildHeader(indx: TDefValue);
var
 s, i, ii          : TDefValue;
 curSize, count    : TAsmValue;
 iFnd              : boolean;
 id                : TDefValue;
 vt1, opr          : byte;
 val1, val2, siz   : TAsmValue;
 str               : string;
 value, blockSiz   : TAsmValue;
 tmp, segOfs       : TAsmValue;
 segI              : TDefValue;
begin
   EmitLn('------ START BUILDING HEADER [' + MakeTime(true) + '] ------');
   WrittenBytes:= HeaderList[indx].HeaderSize;
   if SegCount > 0 then begin
      s:= 0;
      while s < SegCount do begin
         if Segment[s].WriteSEG then begin
            WrittenBytes:= WrittenBytes + Segment[s].Offset;
         end;
         Inc(s);
      end;
   end;
   if OutputList[CurOutput].FillSize > 0 then begin
      if WrittenBytes < OutputList[CurOutput].FillSize then
       WrittenBytes:= OutputList[CurOutput].FillSize;
   end;
   if OutputList[CurOutput].RoundSize > 0 then begin
      curSize:= HeaderList[indx].HeaderSize;
      count:= curSize div OutputList[CurOutput].RoundSize;
      count:= curSize - (count * OutputList[CurOutput].RoundSize);
      if (count > 0) and (count < OutputList[CurOutput].RoundSize) then begin
         count:= OutputList[CurOutput].RoundSize - count;
         WrittenBytes:= WrittenBytes + count;
      end;
   end;

   if NewHeaderItemCount > 0 then begin
      i:= 0;
      while i < NewHeaderItemCount do begin
         IsErr:= false;
         id:= NewHeaderItemList[i].ID;
         vt1:= NewHeaderItemList[i].Type1;
         {vt2:= NewHeaderItemList[i].Type2; }
         val1:= NewHeaderItemList[i].Value1;
         val2:= NewHeaderItemList[i].Value2;
         opr:= NewHeaderItemList[i].Opr;
         str:= NewHeaderItemList[i].StrValue;
         iFnd:= false;
    	   if (id = hdrVALUE) and (not iFnd) then begin
            iFnd:= true;
            ValueToBytes(val1);
            siz:= 1;
            if vt1 = tWORD then siz:= 2;
            if vt1 = tDWORD then siz:= 4;
            if vt1 = tQWORD then siz:= 8;
            if vt1 = tFARPTR then siz:= 6;
            if vt1 = tTENBYTE then siz:= 10;
            SetLength(OutputMem, OutputLen + siz);
            ii:= 0;
            while ii < siz do begin
               OutputMem[OutputLen+ii]:= byt[ii+1];
               Inc(ii);
            end;
            OutputLen:= OutputLen + siz;
         end;
         if (id = hdrMEMOWRITE) and (not iFnd) then begin
            iFnd:= true;
            value:= hdrMem[val1];
            if opr <> oprNONE then begin
               value:= DoCalc(hdrMem[val1], opr, hdrMem[val2]);
            end;
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
         if (id = hdrBLOCKCNT) and (not iFnd) then begin
            iFnd:= true;
            value:= HeaderList[indx].HeaderSize + WrittenBytes;
            blockSiz:= OutputList[CurOutput].BlockSize;
            value:= (value div blockSiz) + 1;
            if opr <> oprNONE then value:= DoCalc(value, opr, val1);
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
         if (id = hdrLASTBLOCKSIZE) and (not iFnd) then begin
            iFnd:= true;
            value:= HeaderList[indx].HeaderSize + WrittenBytes;
            blockSiz:= OutputList[CurOutput].BlockSize;
            tmp:= value div blockSiz;
            value:= value - (tmp * blockSiz);
            if opr <> oprNONE then value:= DoCalc(value, opr, val1);
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
         if (id = hdrRELOCCNT) and (not iFnd) then begin
            iFnd:= true;
            value:= RelocCount;
            if opr <> oprNONE then value:= DoCalc(value, opr, val1);
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
         if (id = hdrSIZE) and (not iFnd) then begin
            iFnd:= true;
            value:= HeaderList[indx].HeaderSize;
            if opr <> oprNONE then value:= DoCalc(value, opr, val1);
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
         if (id = hdrSTACK) and (not iFnd) then begin
            iFnd:= true;
            value:= StackMem;
            if opr <> oprNONE then value:= DoCalc(value, opr, val1);
            if not IsErr then begin
               ValueToBytes(value);
	             siz:= 1;
               if vt1 = tWORD then siz:= 2;
               if vt1 = tDWORD then siz:= 4;
               if vt1 = tQWORD then siz:= 8;
               if vt1 = tFARPTR then siz:= 6;
               if vt1 = tTENBYTE then siz:= 10;
               SetLength(OutputMem, OutputLen + siz);
               ii:= 0;
               while ii < siz do begin
                  OutputMem[OutputLen+ii]:= byt[ii+1];
                  Inc(ii);
               end;
               OutputLen:= OutputLen + siz;
            end;
         end;
	       if (id = hdrSEGOFFSET) and (not iFnd) then begin
            iFnd:= true;
            segI:= Neg;
            segOfs:= 0;
            if SegCount > 0 then begin
               s:= 0;
               while s < SegCount do begin
                  if Segment[s].IdentName = str then begin
                     segI:= s;
                     break;
                  end;
                  segOfs:= segOfs + Segment[s].Offset;
                  Inc(s);
               end;
            end;
            if segI <> Neg then begin
               value:= segOfs;
               if opr <> oprNONE then value:= DoCalc(value, opr, val1);
               if not IsErr then begin
                  ValueToBytes(value);
	                siz:= 1;
                  if vt1 = tWORD then siz:= 2;
                  if vt1 = tDWORD then siz:= 4;
                  if vt1 = tQWORD then siz:= 8;
                  if vt1 = tFARPTR then siz:= 6;
                  if vt1 = tTENBYTE then siz:= 10;
                  SetLength(OutputMem, OutputLen + siz);
                  ii:= 0;
                  while ii < siz do begin
                     OutputMem[OutputLen+ii]:= byt[ii+1];
                     Inc(ii);
                  end;
                  OutputLen:= OutputLen + siz;
               end;
	          end else begin
               FatalError('Segment not found [' + str + ']!');
            end;
         end;
         if (id = hdrSEGSIZE) and (not iFnd) then begin
            iFnd:= true;
            segI:= Neg;
            if SegCount > 0 then begin
               s:= 0;
               while s < SegCount do begin
                  if Segment[s].IdentName = str then begin
                     segI:= s;
                     break;
                  end;
                  Inc(s);
               end;
            end;
            if segI <> Neg then begin
               value:= Segment[segI].Offset;
               if opr <> oprNONE then value:= DoCalc(value, opr, val1);
               if not IsErr then begin
                  ValueToBytes(value);
	                siz:= 1;
                  if vt1 = tWORD then siz:= 2;
                  if vt1 = tDWORD then siz:= 4;
                  if vt1 = tQWORD then siz:= 8;
                  if vt1 = tFARPTR then siz:= 6;
                  if vt1 = tTENBYTE then siz:= 10;
                  SetLength(OutputMem, OutputLen + siz);
                  ii:= 0;
                  while ii < siz do begin
                     OutputMem[OutputLen+ii]:= byt[ii+1];
                     Inc(ii);
                  end;
                  OutputLen:= OutputLen + siz;
               end;
	          end else begin
               FatalError('Segment not found [' + str + ']!');
            end;
         end;
         if not iFnd then begin
            FatalError('Invalid Header Item ID [' + IntToStr(id) + ']!');
         end;
         Inc(i);
      end;
   end;
   EmitLn('------ END BUILDING HEADER [' + MakeTime(true) + '] ------');
end;

end.
