{ Main - Init & Starting the assembler 
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _main;

interface

uses
 _common;

var
 IsRunning         : boolean;
 StopASM           : boolean;

 AsmFile           : string;
 IDataFile         : string;
 EDataFile         : string;
 OutFile           : string;

 TERMFound         : boolean;
 ENDFound          : boolean;
 EXITFound         : boolean;

 needEquUpdate     : boolean;

procedure InitASM;
procedure StartASM(afi, ifi, efi, ofi: string);

procedure AsmLine;

procedure CheckEndLine;
procedure CheckStructEnds;
procedure CheckParam(parName: string;
 parOfs: TDefValue; parCnt: TAsmValue);

function GrabParams(prmOfs: TDefValue; isObj: boolean): TAsmValue;

procedure MakeMacro(macName: string);
procedure MakeObject(objName: string);
function GrabParamValues(prmOfs: TDefValue): TAsmValue;
function GrabObjectParamValues(prmOfs: TDefValue): TAsmValue;

function GetMacroParam(n: string;
 po: TDefValue; pc: TAsmValue;
 pvo: TDefValue; pvc: TAsmValue): string;
function InsertParams(s: string): string;

procedure BuildOutput;

implementation

uses
 SysUtils, _emit, _func, _mem, _error, _alias, _ident,
 _codegen, _source, _proc, _macro, _sec, _seg, _conv,
 _reg, _cpu, _parser, _model, _header, _output,
 _modrm, _dx, _equ, _state, _diskio, _scanner, _gen, _pie,
 _mainsrc;


procedure InitASM;
begin
   InitMemData;
   InitErrors;
   InitAlias;
   InitIdents;
   InitCodeLists;
   InitSource;
   InitProc;
   InitMacros;
   InitSections;
   InitSegments;
   InitRegs;
   InitProcessors;
   InitModels;
   InitHeaders;
   InitOutputs;
   InitModRM;
   InitGenerator;
   InitReport;
   InitInstr;

   PriName:= '';
   SubName:= '';

   retOpr:= oprNONE;

   TERMFound:= false;
   ENDFound:= false;
   EXITFound:= false;
   needEquUpdate:= false;
   StopASM:= false;
end;

procedure StartASM(afi, ifi, efi, ofi: string);
var
 useIData          : boolean;
 useEData, useEsrc : boolean;
 sOff              : TDefValue;
 nLin              : TAsmValue;
 lv                : TDefValue;
 s                 : string;
 doBuild           : boolean;
begin
   InitASM;

   AsmFile:= afi;
   IDataFile:= ifi;
   EDataFile:= efi;
   OutFile:= ofi;

   EmitLn('--- START ASSEMBLING [' + MakeTime(true) + '] ---');
   if AsmFile <> '' then begin
      if not FileExists(AsmFile) then begin
         if FileExists(AsmFile + '.asm') then begin
            AsmFile:= AsmFile + '.asm';
         end else begin
            SysError('[' + AsmFile + '] not found!');
         end;
      end;
      if not IsErr then begin
         if (AsmFile <> IDataFile) and (AsmFile <> EDataFile) then begin
            if OutFile = '' then OutFile:= RemoveExt(AsmFile);

            useIData:= false;
            useEData:= false;
            useEsrc:= false;
            if IDataFile <> '' then begin
               useIData:= true;
               Inc(CurID);
               SetSource(0, CurID, IDataFile, srcFILE);
               ReadSource(0);
               sOff:= Source[0].SrcOffset;
               nLin:= Source[0].NumLines;
               AddSourceData(0, srcFILE, sOff, nLin, IDataFile);
            end else begin
               {Inc(CurID);
               SetSource(0, CurID, AsmFile, srcFILE);
               ReadSource(0);
               sOff:= Source[0].SrcOffset;
               nLin:= Source[0].NumLines;
               AddSourceData(0, srcFILE, sOff, nLin, AsmFile);}
               useIData:= true;
               Inc(CurID);
               SetSource(0, CurID, IDataFile, srcFILE);
               GetInitialData(0);
               sOff:= Source[0].SrcOffset;
               nLin:= Source[0].NumLines;
               AddSourceData(0, srcFILE, sOff, nLin, IDataFile);
            end;
            if EDataFile <> '' then
             useEData:= true
            else
             useEsrc:= true;

            AddStandardEqus;

            while true do begin

               {if BuildConfig.DispLine then begin
                  frmBuild.Label16.Caption:= '[' + Source[CurLevel].FileName + ']';
                  frmBuild.Label16.Hint:= Source[CurLevel].FileName;

                  frmBuild.Label12.Caption:= '[' + Source[CurLevel].SrcLine + ']';
                  frmBuild.Label12.Hint:= Source[CurLevel].SrcLine;

                  cL:= Source[CurLevel].CurLine;
                  tL:= Source[CurLevel].NumLines;
                  if tL > 0 then begin
                     if cL <= tL then
                      frmBuild.Label13.Caption:= IntToStr(cL) + ' of ' + IntToStr(tL)
                     else
                      frmBuild.Label13.Caption:= IntToStr(tL) + ' of ' + IntToStr(tL);
                  end else begin
                     frmBuild.Label13.Caption:= '0 of 0';
                  end;
                  frmBuild.Label16.Update;
                  frmBuild.Label12.Update;
                  frmBuild.Label13.Update;
               end;}

               if StopASM then break;

               if TERMFound then break;

               if ENDFound then begin
                  Source[FileLevel].CurLine:= Source[FileLevel].NumLines + 1;
                  CurLevel:= FileLevel;
                  ENDFound:= false;
               end;

               if EXITFound then begin
                  Source[CurLevel].CurLine:= Source[CurLevel].NumLines + 1;
                  EXITFound:= false;
               end;

               if Source[CurLevel].CurLine > Source[CurLevel].NumLines then begin
                  if CurLevel = 0 then begin
                     if useIData then begin
                        UpdateCodeID;
                        CheckStateEnds;
                        CheckStructEnds;
                        useIData:= false;
                        Inc(CurID);
                        SetSource(0, CurID, AsmFile, srcFILE);
                        ReadSource(0);
                        CurLevel:= 0;
                        FileLevel:= 0;
                        ParamValCount:= 0;
                        sOff:= Source[0].SrcOffset;
                        nLin:= Source[0].NumLines;
                        AddSourceData(0, srcFILE, sOff, nLin, AsmFile);
                        continue;
                     end else begin
                        if useEData then begin
                           UpdateCodeID;
                           CheckStateEnds;
                           CheckStructEnds;
                           useEData:= false;
                           Inc(CurID);
                           SetSource(0, CurID, EDataFile, srcFILE);
                           ReadSource(0);
                           CurLevel:= 0;
                           FileLevel:= 0;
                           ParamValCount:= 0;
                           sOff:= Source[0].SrcOffset;
                           nLin:= Source[0].NumLines;
                           AddSourceData(0, srcFILE, sOff, nLin, EDataFile);
                           continue;
                        end else begin
                           if not useEsrc then begin
                              CheckStateEnds;
                              CheckStructEnds;
                              break;
                           end else begin
                              UpdateCodeID;
                              CheckStateEnds;
                              CheckStructEnds;
                              useEsrc:= false;
                              Inc(CurID);
                              SetSource(0, CurID, EDataFile, srcFILE);
                              GetEndData(0);
                              CurLevel:= 0;
                              FileLevel:= 0;
                              ParamValCount:= 0;
                              sOff:= Source[0].SrcOffset;
                              nLin:= Source[0].NumLines;
                              AddSourceData(0, srcFILE, sOff, nLin, EDataFile);
                              continue;
                           end;
                        end;
                     end;
                  end else begin
                     UpdateCodeID;
                     CheckStateEnds;
                     { CheckStructEnds; }
                     ParamValCount:= Source[CurLevel].ParamValOfs;
                     Dec(CurLevel);
                     lv:= CurLevel;
                     while lv >= 0 do begin
                        if Source[lv].SrcType = srcFILE then begin
                           FileLevel:= lv;
                           break;
                        end;
                        Dec(lv);
                     end;
                     continue;
                  end;
               end;

               ENDFound:= false;
               EXITFound:= false;
               IsErr:= false;

               Source[CurLevel].SrcLine:=
                SourceLines[Source[CurLevel].SrcOffset + (Source[CurLevel].CurLine-1)];
               Source[CurLevel].Pos:= 1;
               s:= Source[CurLevel].SrcLine;
               Source[CurLevel].Len:= Length(s);

               if CurLevel > 0 then begin
                  if Source[CurLevel].SrcType = srcMACRO then begin
                     Source[CurLevel].SrcLine:= InsertParams(s);
                     s:= Source[CurLevel].SrcLine;
                     Source[CurLevel].Len:= Length(s);
                  end;
               end;
               UpdateEquates;
               AsmLine;
               Inc(Source[CurLevel].CurLine);
            end;
         end else begin
            if AsmFile = IDataFile then
             SysError('Input filename and Initial Data filename is matching!')
            else
             SysError('Input filename and End Data filename is matching!');
         end;
      end;
   end else begin
      if not IsSysErr then SysError('Input file expected!');
   end;
   UpdateCodeID;
   CloseAllFileObj;
   {if BuildConfig.DispLine then begin
      frmBuild.Label16.Hint:= frmBuild.RemBrackets(frmBuild.Label16.Caption);
      frmBuild.Label12.Hint:= frmBuild.RemBrackets(frmBuild.Label12.Caption);
   end;}
   Emitln('Errors  : ' + IntToStr(ErrorCount));
   Emitln('Warnings: ' + IntToStr(WarningCount));
   EmitLn('--- END ASSEMBLING [' + MakeTime(true) + '] ---');
   EmitLn('');
   if (not IsSysErr) and (not StopASM) then begin
      doBuild:= true;
      if ErrorCount > 0 then doBuild:= false;
      if doBuild then begin
         EmitLn('--- START BUILDING [' + MakeTime(true) + '] ---');
         BuildOutput;
         EmitLn('--- END BUILDING [' + MakeTime(true) + '] ---');
         EmitLn('');
      end;
   end;
end;



procedure AsmLine;
var
 filp              : file of byte;
 tmpBuf            : array[0..1024] of byte;
 TempPos           : integer;
 ModMacro          : boolean;
 State             : boolean;
 Lv, id, i, sI     : TDefValue;
 PriFnd, SubFnd    : boolean;
 Lbl, s, n, ss, m  : string;
 uSeg              : TDefValue;
 uSegName          : string;
 aName, tName      : string;
 r                 : string;
 cod, rVal         : TAsmValue;
 val               : TAsmValue;
 fiOfs             : TDefValue;
 fiLen             : TAsmValue;
 pat               : string;
 hdrName, item     : string;
 iFnd              : boolean;
 vaTyp             : byte;
 vstr              : string;
 c                 : char;
 mem1, mem2        : TAsmValue;
 memOpr, mTyp      : byte;
 sTyp, sOpr        : byte;
 segN              : string;
 wTyp              : byte;
 vOpr              : byte;
 itName            : string;
 hitem             : byte;
 fExt, hdrIdent    : string;
 hdrIndex          : TDefValue;
 bSize             : TAsmValue;
 oLimit            : TAsmValue;
 oFill             : TAsmValue;
 oRound            : TAsmValue;
 executeCmd        : boolean;
 pOfs              : TDefValue;
 pCnt              : TAsmValue;
 mainParam         : string;
 segI              : TDefValue;
 procI             : TDefValue;
 outI              : TDefValue;
 sOff              : TDefValue;
 nLin              : TAsmValue;
 bc                : TDefValue;
 cnt               : TAsmValue;
 ptnOfs            : TDefValue;
 ptnLen, ptnPos    : TAsmValue;
 b                 : byte;
 curSize           : TAsmValue;
 giFnd             : boolean;
 itemDat           : TDynArrayItem;
 typ               : byte;
 indx              : TDefValue;
 t                 : byte;
 rbyt, fpos, flen  : int64;
begin
   PriName:= GetName;
   SubName:= '';
   TempPos:= Source[CurLevel].Pos;

   ModMacro:= false;
   if (PriName = 'ENDM') and (not ModMacro) then begin
      ModMacro:= true;
      if MacroName <> '' then begin
         if not MacroIsObject then begin
            AddMacro(MacroName, MacroStart, MacroLength, MacroParamOffset, MacroParamCount);
            MacroName:= '';
         end else begin
            ErrorMsg('[ENDO] expected, but [ENDM] found!');
         end;
      end else begin
         ErrorMsg('[ENDM] without [MACRO]!');
      end;
   end else begin
      SubName:= GetName;
      if (SubName = 'ENDM')  and (not ModMacro) then begin
         ModMacro:= true;
         if MacroName <> '' then begin
            if PriName = MacroName then begin
               if not MacroIsObject then begin
                  AddMacro(MacroName, MacroStart, MacroLength, MacroParamOffset, MacroParamCount);
                  MacroName:= '';
               end else begin
                  ErrorMsg('[ENDO] expected, but [ENDM] found!');
               end;
            end else begin
               ErrorMsg('[' + PriName + '] does not match macro name!');
            end;
         end else begin
            ErrorMsg('[ENDM] without [MACRO]!');
         end;
      end;
   end;
   if (PriName = 'ENDO')  and (not ModMacro) then begin
      ModMacro:= true;
      if MacroName <> '' then begin
         if MacroIsObject then begin
            AddObject(MacroName, MacroStart, MacroLength, MacroParamOffset, MacroParamCount);
            MacroName:= '';
         end else begin
            ErrorMsg('[ENDM] expected, but [ENDO] found!');
         end;
      end else begin
         ErrorMsg('[ENDO] without [OBJECT]!');
      end;
   end else begin
      Source[CurLevel].Pos:= TempPos;
      SubName:= GetName;
      if (SubName = 'ENDO')  and (not ModMacro) then begin
         ModMacro:= true;
         if MacroName <> '' then begin
            if PriName = MacroName then begin
               if MacroIsObject then begin
                  AddObject(MacroName, MacroStart, MacroLength, MacroParamOffset, MacroParamCount);
                  MacroName:= '';
               end else begin
                  ErrorMsg('[ENDM] expected, but [ENDO] found!');
               end;
            end else begin
               ErrorMsg('[' + PriName + '] does not match object name!');
            end;
         end else begin
            ErrorMsg('[ENDO] without [OBJECT]!');
         end;
      end;
   end;
   if (MacroName <> '') and (not ModMacro) then begin
      ModMacro:= true;
      Source[CurLevel].Pos:= Source[CurLevel].Len + 1;
      Inc(MacroLength);
   end;

   if not ModMacro then begin
      Source[CurLevel].Pos:= TempPos;
      State:= CheckStatements;
      if (not State) and (not IsErr) then begin
         Lv:= Source[CurLevel].StateLevel;
         if Source[CurLevel].Statements[Lv].Cond then begin
            PriFnd:= false;
            if (PriName = 'ENDM') and (not PriFnd) then begin
               PriFnd:= true;
               ErrorMsg('[ENDM] without [MACRO]!');
            end;
            if (PriName = 'ENDO') and (not PriFnd) then begin
               PriFnd:= true;
               ErrorMsg('[ENDO] without [OBJECT]!');
            end;
            if (PriName = 'ENDP') and (not PriFnd) then begin
               PriFnd:= true;
               if ProcName <> '' then begin
                  if not IsErr then begin
                     AddLabel(ProcName, ProcOffset);
                     ProcName:= '';
                  end;
               end else begin
                  ErrorMsg('[ENDP] without [PROC]!');
               end;
            end;
            if (PriName = 'ENDS') and (not PriFnd) then begin
               PriFnd:= true;
               if SecName <> '' then begin
                  if ProcName = '' then begin
                     if not IsErr then begin
                        AddSection(SecName, SecOffset, SecPos);
                        SecName:= '';
                     end;
                  end else begin
                     ErrorMsg('[ENDP] expected, but [ENDS] found!');
                  end;
               end else begin
                  ErrorMsg('[ENDS] without [SECTION]!');
               end;
            end;

	          if (PriName = 'END') and (not PriFnd) then begin
               PriFnd:= true;
               Lbl:= GetName;
               if Lbl <> '' then begin
                  id:= GetIdentIndex(Lbl);
                  if id <> NotFound then begin
                     if IdentList[id].IdentType <> itLABEL then begin
                        ErrorMsg('Identifier is not a label [' + Lbl + ']!');
                     end;
                  end else begin
                     ErrorMsg('Identifier not found [' + Lbl + ']!');
                  end;
               end;
               if not IsErr then begin
                  ENDFound:= true;
               end;
            end;
            if (PriName = 'TERMINATE') and (not PriFnd) then begin
               PriFnd:= true;
               TERMFound:= true;
               s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     if s = 'NOBUILD' then begin
                        StopASM:= true;
                     end else begin
                        ErrorMsg('Invalid Name [' + s + ']!');
                     end;
                  end;
               end;
            end;
            if ((PriName = 'EXITM') or (PriName = 'EXITO')) and (not PriFnd) then begin
               PriFnd:= true;
               if Source[CurLevel].SrcType = srcMACRO then begin
                  EXITFound:= true;
                  Source[CurLevel].StateLevel:= 0;
                  Source[CurLevel].StateSkip:= 0;
               end else begin
                  ErrorMsg('You must use [EXITM]/[EXITO] inside a [MACRO] or [OBJECT]!');
               end;
            end;

            if ((PriName = 'ECHO') or (PriName = '@ECHO')) and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GrabString;
               if not IsErr then begin
                  EmitLn(s);
               end;
            end;
            if ((PriName = 'ERROR') or (PriName = '@ERROR')) and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GrabString;
               if not IsErr then begin
                  ErrorMsg(s);
               end;
            end;
            if ((PriName = 'WARNING') or (PriName = '@WARNING')) and (not PriFnd) then begin
               PriFnd:= true;
               s:= GrabString;
               if not IsErr then begin
                  WarningMsg(s);
               end;
            end;

            if (PriName = 'LOCAL') and (not PriFnd) then begin
               PriFnd:= true;
               n:= GetName;
               if not IsErr then begin
                  if n <> '' then begin
                     id:= GetIdentIndex(n);
                     if id <> NotFound then begin
                        i:= GetLocal(ExtractIdentName(n));
                        if i = NotFound then begin
                           IdentList[id].SourceID:= Source[CurLevel].SourceID;
                        end;
                     end else begin
                        ErrorMsg('Identifier not found [' + n + ']!');
                     end;
                  end else begin
                     ErrorMsg('Identifier expected!');
                  end;
               end;
            end;

            if (PriName = 'GLOBAL') and (not PriFnd) then begin
               PriFnd:= true;
               n:= GetName;
               if not IsErr then begin
                  if n <> '' then begin
                     id:= GetIdentIndex(n);
                     if id <> NotFound then begin
                        i:= GetGlobal(ExtractIdentName(n));
                        if i = NotFound then begin
                           IdentList[id].SourceID:= 0;
                        end;
                     end else begin
                        ErrorMsg('Identifier not found [' + n + ']!');
                     end;
                  end else begin
                     ErrorMsg('Identifier expected!');
                  end;
               end;
            end;

            if ((PriName = 'LOCALMODE') or (PriName = '.LOCALMODE')) and (not PriFnd) then begin
               PriFnd:= true;
               IdentMode:= imLOCAL;
               AddCodeListLOCAL;
            end;

            if ((PriName = 'GLOBALMODE') or (PriName = '.GLOBALMODE')) and (not PriFnd) then begin
               PriFnd:= true;
               IdentMode:= imGLOBAL;
               AddCodeListGLOBAL;
            end;

            if ((PriName = 'INSTR') or (PriName = '.INSTR')) and (not PriFnd) then begin
               PriFnd:= true;
               GrabInstr;
            end;

            if (PriName = '.FIXINSTRPREFIXES') and (not PriFnd) then begin
               PriFnd:= true;
               FixInstr32;
            end;

            if (PriName = '.REMOVEINSTRPREFIXES') and (not PriFnd) then begin
               PriFnd:= true;
               RemoveInstr32;
            end;

            if (PriName = '.DUMPINSTR') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GrabString;
               if not IsErr then begin
                  if s <> '' then begin
                     DumpInstr(s);
                  end else begin
                     ErrorMsg('Filename expected!');
                  end;
               end;
            end;

            if (PriName = '.RUNINSTR') and (not PriFnd) then begin
               PriFnd:= true;
               n:= GetName;
               if not IsErr then begin
                  if n <> '' then begin
                     RunInstr(n);
                     if (not IsErr) and (not RunInstr2) then begin
                        ErrorMsg('Instruction not found [' + n + '] [.RUNINSTR]!');
                     end;
                  end else begin
                     ErrorMsg('Instruction Name expected [.RUNINSTR]!');
                  end;
               end;
            end;

            if ((PriName = 'INCLUDEBIN') or (PriName = '.INCLUDEBIN')) and (not PriFnd) then begin
               PriFnd:= true;
               s:= GrabString;
               if not IsErr then begin
                  if s <> '' then begin
                     if FileExists(s) then begin
                        fpos:= 0;
                        flen:= 0;
                        SkipBlank;
                        if GetSrcChar = ',' then begin
                           NextChar;
                           fpos:= GetNumber(tQWORD);
                           if not IsErr then begin
                              SkipBlank;
                              if GetSrcChar = ',' then begin
                                 NextChar;
                                 flen:= GetNumber(tQWORD);
                              end;
                           end;
                        end;
                        if not IsErr then begin
                           {$I-}
                           Assign(filp, s);
                           Reset(filp);
                              fiLen:= FileSize(filp);
                              if fpos > 0 then begin
                                 if fpos < fiLen then begin
                                    Seek(filp, fpos);
                                 end else begin
                                    ErrorMsg('Start position out of range [' + PriName+ ']!');
                                 end;
                              end;
                              if flen > 0 then begin
                                 flen:= flen + fpos;
                                 if fpos > fiLen then begin
                                    ErrorMsg('Read Length out of range [' + PriName + ']!');
                                 end;
                              end else begin
                                 flen:= fiLen;
                              end;
                              if not IsErr then begin
                                 while fpos < flen do begin
                                    rbyt:= flen - fpos;
                                    if rbyt >= 1024 then begin
                                       BlockRead(filp, tmpBuf[0], 1024);
                                       i:= 0;
                                       while i < 1024 do begin
                                          AddCodeListVALUE(tBYTE, tmpBuf[i]);
                                          Inc(i);
                                       end;
                                       fpos:= fpos + 1024;
                                    end else begin
                                       BlockRead(filp, tmpBuf[0], rbyt);
                                       i:= 0;
                                       while i < rbyt do begin
                                          AddCodeListVALUE(tBYTE, tmpBuf[i]);
                                          Inc(i);
                                       end;
                                       fpos:= fpos + rbyt;
                                    end;
                                 end;
                              end;
                           Close(filp);
                           {$I+}
                        end;
                     end else begin
                        ErrorMsg('File not found [' + s + ']!');
                     end;
                  end else begin
                     ErrorMsg('Filename expected!');
                  end;
               end;
            end;

            if ((PriName = 'INCLUDE') or (PriName = '.INCLUDE')) and (not PriFnd) then begin
               PriFnd:= true;
               s:= GrabString;
               if not IsErr then begin
                  if s <> '' then begin
                     if FileExists(s) then begin
                        CheckEndLine;
                        if not IsErr then begin
                           Inc(Source[CurLevel].CurLine);
                           Inc(CurID);
                           Inc(CurLevel);
                           FileLevel:= CurLevel;
                           SetLength(Source, CurLevel + 1);
                           SetSource(CurLevel, CurID, s, srcFILE);
                           ReadSource(CurLevel);
                           Source[CurLevel].CurLine:= 0;
                           Source[CurLevel].SrcLine:= '';
                           Source[CurLevel].Pos:= Source[CurLevel].Len + 1;
                           sOff:= Source[CurLevel].SrcOffset;
                           nLin:= Source[CurLevel].NumLines;
                           AddSourceData(CurLevel, srcFILE, sOff, nLin, s);
                        end;
                     end else begin
                        ErrorMsg('File not found [' + s + ']!');
                     end;
                  end else begin
                     ErrorMsg('Filename expected!');
                  end;
               end;
            end;

            if ((PriName = 'DELETE') or (PriName = '.DELETE')) and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if not DeleteIdent(s) then begin
                     ErrorMsg('Identifier not found [' + s + ']!');
                  end;
               end;
            end;

	          if (PriName = 'MACRO') and (not PriFnd) then begin
               PriFnd:= true;
               MakeMacro('');
            end;
            if (PriName = 'OBJECT') and (not PriFnd) then begin
               PriFnd:= true;
               MakeObject('');
            end;
            if (PriName = 'PROC') and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if ProcName = '' then begin
                     if s <> '' then begin
                        ProcName:= s;
                        ProcOffset:= GetSegOffset(false);
                     end else begin
                        ErrorMsg('Procedure name expected!');
                     end;
                  end else begin
                     ErrorMsg('Can not create a procedure inside another procedure!');
                  end;
               end;
            end;
            if (PriName = 'SECTION') and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if SecName = '' then begin
                     if ProcName = '' then begin
                        if s <> '' then begin
                           uSegName:= GetName;
                           uSeg:= CurSeg;
                           if uSegName <> '' then begin
                              uSeg:= Neg;
                              i:= 0;
                              while i < segCount do begin
                                 if uSegName = Segment[i].IdentName then begin
                                    uSeg:= i;
                                    break;
                                 end;
                                 Inc(i);
                              end;
                              if uSeg = Neg then begin
                                 ErrorMsg('Segment not found [' + uSegName + ']!');
                              end;
                           end;
                           if not IsErr then begin
                              SecOffset:= GetSegOffset(true);
                              SecName:= s;
                              SecSegment:= uSeg;
                              SecPos:= 0;
                              SecORG:= 0;
                              SecRound:= 0;
                           end;
                        end else begin
                           ErrorMsg('Section name expected!');
                        end;
                     end else begin
                        ErrorMsg('Can not create a section inside a procedure!');
                     end;
                  end else begin
                     ErrorMsg('Can not create a section inside another section!');
                  end;
               end;
            end;

	          if ((PriName = 'ALIAS') or (PriName = '.ALIAS')) and (not PriFnd) then begin
               PriFnd:= true;
               aName:= GetName;
               if aName <> '' then begin
                  tName:= GetName;
                  if tName <> '' then begin
                     if not IsErr then begin
                        AddAlias(aName, tName);
                     end;
                  end else begin
                     ErrorMsg('Target name expected [ALIAS]!');
                  end;
               end else begin
                  ErrorMsg('Name expected [ALIAS]!');
               end;
            end;

            if ((PriName = 'ALIASCHAR') or (PriName = '.ALIASCHAR')) and (not PriFnd) then begin
               PriFnd:= true;
               aName:= GrabString;
               if not IsErr then begin
                  tName:= GrabString;
               end;
               if not IsErr then begin
                  if Length(aName) = 1 then begin
                     if Length(tName) = 1 then begin
                        AddAliasChar(aName[1], tName[1]);
                     end else begin
                        if Length(tName) = 0 then begin
                           ErrorMsg('Character expected in string (Target Character) [ALIASCHAR]!');
                        end else begin
                           ErrorMsg('Only one character allowed in string (Target Character) [ALIASCHAR]!');
                        end;
                     end;
                  end else begin
                     if Length(aName) = 0 then begin
                        ErrorMsg('Character expected in string (Alias Character) [ALIASCHAR]!');
                     end else begin
                        ErrorMsg('Only one character allowed in string (Alias Character) [ALIASCHAR]!');
                     end;
                  end;
               end;
            end;

            if (PriName = '.REG') and (not PriFnd) then begin
               PriFnd:= true;
               r:= GetName;
               if not IsErr then begin
                  if r <> '' then begin
                     cod:= GetNumber(tAUTO);
                     if not IsErr then begin
                        AddReg(r, cod, 0);
                     end;
                  end else begin
                     ErrorMsg('Register name expected!');
                  end;
               end;
            end;
            if (PriName = '.SEGREG') and (not PriFnd) then begin
               PriFnd:= true;
               r:= GetName;
               if not IsErr then begin
                  if r <> '' then begin
                     cod:= GetNumber(tAUTO);
                     if not IsErr then begin
                        rVal:= GetNumber(tAUTO);
                        if not IsErr then begin
                           AddReg(r, cod, rVal);
                        end;
                     end;
                  end else begin
                     ErrorMsg('Register name expected!');
                  end;
               end;
            end;

            if (PriName = '.REGTABLE') and (not PriFnd) then begin
               PriFnd:= true;
               r:= GetName;
               if not IsErr then begin
                  if r <> '' then begin
                     AddRegTable(r);
                  end else begin
                     ErrorMsg('Name expected!');
                  end;
               end;
            end;

            if (PriName = '.MODRMTBL') and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        AddModRMTable(s, val);
                     end;
                  end else begin
                     ErrorMsg('ModR/M table name expected!');
                  end;
               end;
            end;
            if (PriName = '.MODRMDATA') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     AddModRMData(s);
                  end else begin
                     ErrorMsg('ModR/M table name expected!');
                  end;
               end;
            end;

            if ((PriName = 'SEGMENT') or (PriName = '.SEGMENT')) and (not PriFnd) then begin
               PriFnd:= true;
               aName:= GrabString;
               if not IsErr then begin
                  s:= GetName;
                  if not IsErr then begin
                     if s <> '' then begin
                        fiOfs:= 0;
                        fiLen:= 0;
                        pat:= GetName;
                        if not IsErr then begin
                           if pat <> '' then begin
                              id:= GetIdentIndex(pat);
                              if id <> NotFound then begin
                                 if IdentList[id].IdentType = itEQU then begin
                                    if IdentList[id].EquMode = equModeSTR then begin
                                       fiOfs:= IdentList[id].SValue;
                                       fiLen:= IdentList[id].StrLen;
                                    end else begin
                                       ErrorMsg('Expected string equate [' + pat + ']!');
                                    end;
                                 end else begin
                                    ErrorMsg('Expected string equate [' + pat + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Identifier not found [' + pat + ']!');
                              end;
                           end;
                           if not IsErr then begin
                              AddSegment(aName, s, fiOfs, fiLen);
                           end;
                        end;
                     end else begin
                        ErrorMsg('Segment identifier expected!');
                     end;
                  end;
               end;
            end;

            if (PriName = '.LIMIT') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     Segment[CurSeg].Limit:= val;
                  end;
               end else begin
                  ErrorMsg('No segments exists [.LIMIT]!');
               end;
            end;
            if (PriName = '.ROUND') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     cnt:= 0;
                     if val > 0 then begin
                        curSize:= Segment[CurSeg].Offset;
                        cnt:= curSize div val;
                        cnt:= curSize - (cnt * val);
                        if (cnt > 0) and (cnt < val) then begin
                           cnt:= val - cnt;
                        end;
                     end;
                     ptnOfs:= Segment[CurSeg].FillPatternOfs;
                     ptnLen:= Segment[CurSeg].FillPatternLen;
                     ptnPos:= Segment[CurSeg].FillPatternPos;
                     bc:= 1;
                     while bc <= cnt do begin
                        b:= 0;
                        if ptnLen > 0 then begin
                           b:= MemData[ptnOfs + ptnPos];
                           Inc(ptnPos);
                           if ptnPos > (ptnLen-1) then ptnPos:= 0;
                        end;
                        AddCodeListValue(tBYTE, b);
                        Inc(bc);
                     end;
                     Segment[CurSeg].FillPatternPos:= ptnPos;
                  end;
               end else begin
                  ErrorMsg('No segments exists [.ROUND]!');
               end;
            end;
            if (PriName = '.FILL') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     cnt:= 0;
                     if val > 0 then begin
                        if Segment[CurSeg].Offset < val then begin
                           cnt:= val - Segment[CurSeg].Offset;
                        end;
                     end;
                     ptnOfs:= Segment[CurSeg].FillPatternOfs;
                     ptnLen:= Segment[CurSeg].FillPatternLen;
                     ptnPos:= Segment[CurSeg].FillPatternPos;
                     bc:= 1;
                     while bc <= cnt do begin
                        b:= 0;
                        if ptnLen > 0 then begin
                           b:= MemData[ptnOfs + ptnPos];
                           Inc(ptnPos);
                           if ptnPos > (ptnLen-1) then ptnPos:= 0;
                        end;
                        AddCodeListValue(tBYTE, b);
                        Inc(bc);
                     end;
                     Segment[CurSeg].FillPatternPos:= ptnPos;
                  end;
               end else begin
                  ErrorMsg('No segments exists [.FILL]!');
               end;
            end;

            if (PriName = '.ROUNDSEG') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     Segment[CurSeg].RoundSize:= val;
                  end;
               end else begin
                  ErrorMsg('No segments exists [.ROUNDSEG]!');
               end;
            end;
            if (PriName = '.FILLSEG') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     Segment[CurSeg].FillSize:= val;
                  end;
               end else begin
                  ErrorMsg('No segments exists [.FILLSEG]!');
               end;
            end;
            if (PriName = '.RESETFILLPATTERNPOS') and (not PriFnd) then begin
               PriFnd:= true;
               if CurSeg <> Neg then begin
                  Segment[CurSeg].FillPatternPos:= 0;
               end else begin
                  ErrorMsg('No segments exists [.RESETFILLPATTERNPOS]!');
               end;
            end;
	          { Case ".RESETSEG":
              PriFnd:= true;
              // To Do: Reset Segment
              Break;
            }

	          if (PriName = '.ADDRELOC') and (not PriFnd) then begin
               PriFnd:= true;
               val:= GetNumber(tAUTO);
               if not IsErr then begin
                  AddReloc(val);
               end;
            end;

            if (PriName = '.HEADER') and (not PriFnd) then begin
               PriFnd:= true;
               aName:= GrabString;
               if not IsErr then begin
                  s:= GetName;
                  if not IsErr then begin
                     if s <> '' then begin
                        fiOfs:= 0;
                        fiLen:= 0;
                        pat:= GetName;
                        if not IsErr then begin
                           if pat <> '' then begin
                              id:= GetIdentIndex(pat);
                              if id <> NotFound then begin
                                 if IdentList[id].IdentType = itEQU then begin
                                    if IdentList[id].EquMode = equModeSTR then begin
                                       fiOfs:= IdentList[id].SValue;
                                       fiLen:= IdentList[id].StrLen;
                                    end else begin
                                       ErrorMsg('Expected string equate [' + pat + ']!');
                                    end;
                                 end else begin
                                    ErrorMsg('Expected string equate [' + pat + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Identifier not found [' + pat + ']!');
                              end;
                           end;
                           if not IsErr then begin
                              AddHeader(aName, s, fiOfs, fiLen);
                           end;
                        end;
                     end else begin
                        ErrorMsg('Header identifier expected!');
                     end;
                  end;
               end;
            end;

            if (PriName = '.HEADERITEM') and (not PriFnd) then begin
               PriFnd:= true;
               hdrName:= GetName;
               if not IsErr then begin
                  if hdrName <> '' then begin
                     item:= GetName;
                     if not IsErr then begin
                        if item <> '' then begin
                           iFnd:= false;
                           if (item = 'TEXT') and (not iFnd) then begin
                              iFnd:= true;
			                        s:= GrabString;
                              if not IsErr then begin
                                 AddHeaderText(hdrName, s);
                              end;
                           end;
                           if (item = 'BYTE') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tBYTE);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tBYTE, val);
                              end;
                           end;
                           if (item = 'WORD') and (not iFnd) then begin
			                        iFnd:= true;
                              val:= GetNumber(tWORD);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tWORD, val);
                              end;
                           end;
                           if (item = 'DWORD') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tDWORD);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tDWORD, val);
                              end;
                           end;
                           if (item = 'QWORD') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tAUTO);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tQWORD, val);
                              end;
                           end;
                           if (item = 'FARPTR') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tAUTO);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tFARPTR, val);
                              end;
                           end;
                           if (item = 'TENBYTE') and (not iFnd) then begin
                              iFnd:= true;
			                        val:= GetNumber(tAUTO);
                              if not IsErr then begin
                                 AddHeaderValue(hdrName, tTENBYTE, val);
                              end;
                           end;

                           if (item = 'RELOCTBL') and (not iFnd) then begin
                              iFnd:= true;
                              vaTyp:= tNONE;
                              vstr:= GetName;
                              if not IsErr then begin
                                 if vstr = 'BYTE' then vaTyp:= tBYTE;
                                 if vstr = 'WORD' then vaTyp:= tWORD;
                                 if vstr = 'DWORD' then vaTyp:= tDWORD;
                                 if vstr = 'QWORD' then vaTyp:= tQWORD;
                                 if vstr = 'FARPTR' then vaTyp:= tFARPTR;
                                 if vstr = 'TENBYTE' then vaTyp:= tTENBYTE;
                                 if vaTyp <> tNONE then begin
                                    AddHeaderRelocTable(hdrName, vaTyp);
                                 end else begin
                                    if vstr <> '' then begin
                                       ErrorMsg('Invalid RELOCATION type [' + vstr + ']!');
                                    end else begin
                                       ErrorMsg('RELOCATION type expected!');
                                    end;
                                 end;
                              end;
                           end;

                           if (item = 'ROUND') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tAUTO);
                              if not IsErr then begin
                                 AddHeaderRound(hdrName, val);
                              end;
                           end;
                           if (item = 'FILL') and (not iFnd) then begin
                              iFnd:= true;
                              val:= GetNumber(tAUTO);
                              if not IsErr then begin
                                 AddHeaderFill(hdrName, val);
                              end;
                           end;

			                     if (item = 'MEMO') and (not iFnd) then begin
                              iFnd:= true;
                              SkipBlank;
                              if IsAlpha(GetSrcChar) then begin
                                 c:= UpCase(GetSrcChar);
                                 NextChar;
                                 AddHeaderMemo(hdrName, (ord(c) - ord('A')));
                              end else begin
                                 ErrorMsg('Header variable [A]-[Z] expected!');
                              end;
                           end;
                           if (item = 'WRITEMEMO') and (not iFnd) then begin
                              iFnd:= true;
                              mem1:= 0;
                              mem2:= 0;
                              memOpr:= oprNONE;
                              mTyp:= tNONE;
                              tName:= GetName;
                              if not IsErr then begin
                                 if tName <> '' then begin
                                    if tName = 'BYTE' then mTyp:= tBYTE;
                                    if tName = 'WORD' then mTyp:= tWORD;
                                    if tName = 'DWORD' then mTyp:= tDWORD;
                                    if tName = 'QWORD' then mTyp:= tQWORD;
                                    if tName = 'FARPTR' then mTyp:= tFARPTR;
                                    if tName = 'TENBYTE' then mTyp:= tTENBYTE;
                                    if mTyp = tNONE then
                                     ErrorMsg('Invalid type identifier [' + tName + ']!');
                                    if not IsErr then begin
                                       SkipBlank;
                                       if IsAlpha(GetSrcChar) then begin
                                          c:= UpCase(GetSrcChar);
                                          NextChar;
                                          mem1:= ord(c) - ord('A');
                                          SkipBlank;
                                          if IsStdOpr(GetSrcChar) then begin
                                             if GetSrcChar = '+' then memOpr:= oprADD;
                                             if GetSrcChar = '-' then memOpr:= oprSUB;
                                             if GetSrcChar = '*' then memOpr:= oprMUL;
                                             if GetSrcChar = '/' then memOpr:= oprDIV;
                                             NextChar;
                                             SkipBlank;
                                             if IsAlpha(GetSrcChar) then begin
                                                c:= UpCase(GetSrcChar);
                                                NextChar;
                                                mem2:= ord(c) - ord('A');
                                             end else begin
                                                ErrorMsg('Header variable [A]-[Z] expected!');
                                             end;
                                          end;
                                          if not IsErr then begin
                                             AddHeaderMemoWrite(hdrName, mTyp, mem1, memOpr, mem2);
                                          end;
                                       end else begin
                                          ErrorMsg('Header variable [A]-[Z] expected!');
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Type identifier expected!');
                                 end;
                              end;
                           end;

			                     if (item = 'GETSEGSIZE') and (not iFnd) then begin
                              iFnd:= true;
                              sTyp:= tNONE;
                              sOpr:= oprNONE;
                              {val:= 0;}
                              segN:= '';
                              tName:= GetName;
                              if not IsErr then begin
                                 if tName <> '' then begin
                                    if tName = 'BYTE' then sTyp:= tBYTE;
                                    if tName = 'WORD' then sTyp:= tWORD;
                                    if tName = 'DWORD' then sTyp:= tDWORD;
                                    if tName = 'QWORD' then sTyp:= tQWORD;
                                    if tName = 'FARPTR' then sTyp:= tFARPTR;
                                    if tName = 'TENBYTE' then sTyp:= tTENBYTE;
                                    if sTyp = tNONE then
                                     ErrorMsg('Invalid type identifier [' + tName + ']!');
                                    if not IsErr then begin
                                       segN:= GetName;
                                       if not IsErr then begin
                                          if segN <> '' then begin
                                             val:= GetCalc;
                                             if not IsErr then begin
                                                sOpr:= RetOpr;
                                                AddHeaderSegSize(hdrName, sTyp, segN, sOpr, val);
                                             end;
                                          end else begin
                                             ErrorMsg('Segment name expected!');
                                          end;
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Type identifier expected!');
                                 end;
                              end;
                           end;
                           if (item = 'GETSEGOFFSET') and (not iFnd) then begin
                              iFnd:= true;
                              sTyp:= tNONE;
                              sOpr:= oprNONE;
                              { val:= 0; }
                              segN:= '';
                              tName:= GetName;
                              if not IsErr then begin
                                 if tName <> '' then begin
                                    if tName = 'BYTE' then sTyp:= tBYTE;
                                    if tName = 'WORD' then sTyp:= tWORD;
                                    if tName = 'DWORD' then sTyp:= tDWORD;
                                    if tName = 'QWORD' then sTyp:= tQWORD;
                                    if tName = 'FARPTR' then sTyp:= tFARPTR;
                                    if tName = 'TENBYTE' then sTyp:= tTENBYTE;
                                    if sTyp = tNONE then
                                     ErrorMsg('Invalid type identifier [' + tName + ']!');
                                    if not IsErr then begin
                                       segN:= GetName;
                                       if not IsErr then begin
                                          if segN <> '' then begin
                                             val:= GetCalc;
                                             if not IsErr then begin
                                                sOpr:= RetOpr;
                                                AddHeaderSegOffset(hdrName, sTyp, segN, sOpr, val);
                                             end;
                                          end else begin
                                             ErrorMsg('Segment name expected!');
                                          end;
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Type identifier expected!');
                                 end;
                              end;
                           end;

                           if (item = 'WRITE') and (not iFnd) then begin
                              iFnd:= true;
                              wTyp:= tNONE;
                              { val:= 0; }
                              vOpr:= oprNONE;
                              tName:= GetName;
                              if not IsErr then begin
                                 if tName <> '' then begin
				                            if tName = 'BYTE' then wTyp:= tBYTE;
                                    if tName = 'WORD' then wTyp:= tWORD;
                                    if tName = 'DWORD' then wTyp:= tDWORD;
                                    if tName = 'QWORD' then wTyp:= tQWORD;
                                    if tName = 'FARPTR' then wTyp:= tFARPTR;
                                    if tName = 'TENBYTE' then wTyp:= tTENBYTE;
                                    if wTyp = tNONE then
                                     ErrorMsg('Invalid type identifier [' + tName + ']!');
                                    if not IsErr then begin
                                       itName:= GetName;
                                       if not IsErr then begin
                                          if itName <> '' then begin
                                             hitem:= hdrNONE;
                                             if itName = 'BLOCKCNT' then hitem:= hdrBLOCKCNT;
                                             if itName = 'LASTBLOCKSIZE' then hitem:= hdrLASTBLOCKSIZE;
                                             if itName = 'RELOCCNT' then hitem:= hdrRELOCCNT;
                                             if itName = 'SIZE' then hitem:= hdrSIZE;
                                             if itName = 'STACK' then hitem:= hdrSTACK;
                                             if hitem <> hdrNONE then begin
                                                val:= GetCalc;
                                                if not IsErr then begin
                                                   vOpr:= RetOpr;
                                                   AddHeaderWriteItem(hdrName, wTyp, hitem, vOpr, val);
                                                end;
                                             end else begin
                                                ErrorMsg('Invalid item name [' + itName + ']!');
                                             end;
                                          end else begin
                                             ErrorMsg('Item name expected!');
                                          end;
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Type identifier expected!');
                                 end;
                              end;
                           end;

                           if not iFnd then begin
                              ErrorMsg('Invalid header item [' + item + ']!');
                           end;
                        end else begin
                           ErrorMsg('Header item expected!');
                        end;
                     end;
                  end else begin
                     ErrorMsg('Header name expected!');
                  end;
               end;
            end;

            if (PriName = '.OUTPUT') and (not PriFnd) then begin
               PriFnd:= true;
               aName:= GrabString;
               if not IsErr then begin
	                s:= GetName;
                  if not IsErr then begin
                     if s <> '' then begin
                        hdrIdent:= '';
                        hdrIndex:= Neg;
                        bSize:= 0;
                        oLimit:= 0;
                        oFill:= 0;
                        oRound:= 0;
                        fExt:= GrabString;
                        if not IsErr then hdrIdent:= GetName;
                        if not IsErr then begin
                           i:= 0;
                           while i < HeaderCount do begin
                              if hdrIdent = HeaderList[i].IdentName then begin
                                 hdrIndex:= i;
                                 break;
                              end;
                              Inc(i);
                           end;
                           if hdrIndex = Neg then begin
                              ErrorMsg('Header not found [' + hdrIdent + ']!');
                           end;
                        end;
                        if not IsErr then bSize:= GetNumber(tAUTO);
                        if not IsErr then oLimit:= GetNumber(tAUTO);
                        if not IsErr then oFill:= GetNumber(tAUTO);
                        if not IsErr then oRound:= GetNumber(tAUTO);
                        fiOfs:= 0;
                        fiLen:= 0;
                        pat:= '';
                        if not IsErr then pat:= GetName;
                        if not IsErr then begin
                           if pat <> '' then begin
                              id:= GetIdentIndex(pat);
                              if id <> NotFound then begin
                                 if IdentList[id].IdentType = itEQU then begin
                                    if IdentList[id].EquMode = equModeSTR then begin
                                       fiOfs:= IdentList[id].SValue;
                                       fiLen:= IdentList[id].StrLen;
                                    end else begin
                                       ErrorMsg('Expected string equate [' + pat + ']!');
                                    end;
                                 end else begin
                                    ErrorMsg('Expected string equate [' + pat + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Identifier not found [' + pat + ']!');
                              end;
                           end;
                           if not IsErr then begin
                              AddOutput(aName, s, fExt, hdrIndex, bSize, oLimit, oFill, oRound, fiOfs, fiLen);
                           end;
                        end;
                     end else begin
                        ErrorMsg('Output identifier expected!');
                     end;
                  end;
               end;
            end;
            if (PriName = '.OUTLIMIT') and (not PriFnd) then begin
               PriFnd:= true;
               if CurOutput <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     OutputList[CurOutput].Limit:= val;
                  end;
               end else begin
                  ErrorMsg('No output selected [.OUTLIMIT]!');
               end;
            end;
            if (PriName = '.OUTROUND') and (not PriFnd) then begin
               PriFnd:= true;
               if CurOutput <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     OutputList[CurOutput].RoundSize:= val;
                  end;
               end else begin
                  ErrorMsg('No output selected [.OUTROUND]!');
               end;
            end;
            if (PriName = '.OUTFILL') and (not PriFnd) then begin
               PriFnd:= true;
               if CurOutput <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     OutputList[CurOutput].FillSize:= val;
                  end;
               end else begin
                  ErrorMsg('No output selected [.OUTFILL]!');
               end;
            end;
            if (PriName = '.EXT') and (not PriFnd) then begin
               PriFnd:= true;
               if CurOutput <> Neg then begin
	                s:= GrabString;
                  if not IsErr then begin
                     OutputList[CurOutput].FileExt:= s;
                  end;
               end else begin
                  ErrorMsg('No output selected [.EXT]!');
               end;
            end;
            if (PriName = '.BLOCKSIZE') and (not PriFnd) then begin
               PriFnd:= true;
               if CurOutput <> Neg then begin
                  val:= GetNumber(tAUTO);
                  if not IsErr then begin
                     OutputList[CurOutput].BlockSize:= val;
                  end;
               end else begin
                  ErrorMsg('No output selected [.BLOCKSIZE]!');
               end;
            end;

            if (PriName = '.PROCESSOR') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        AddProcessor(s, val);
                     end;
                  end else begin
                     ErrorMsg('Processor name expected!');
                  end;
               end;
            end;

	          if (PriName = '.ADDMODEL') and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     AddModel(s);
                  end else begin
                     ErrorMsg('Model name expected!');
                  end;
               end;
            end;
            if ((PriName = '.MODEL') or (PriName = 'MODEL')) and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     SetModel(s);
                  end else begin
                     ErrorMsg('Model name expected!');
                  end;
               end;
            end;

            if ((PriName = '.ORG') or (PriName = 'ORG')) and (not PriFnd) then begin
               PriFnd:= true;
               val:= GetNumber(tAUTO);
               if not IsErr then begin
	                id:= GetSeg;
                  if id <> Neg then begin
                     Segment[id].ORG:= val;
                  end else begin
                     ErrorMsg('No segment defined, can not use [ORG]!');
                  end;
               end;
            end;

            if ((PriName = '.STACK') or (PriName = 'STACK')) and (not PriFnd) then begin
               PriFnd:= true;
               val:= GetNumber(tAUTO);
               if not IsErr then begin
                  StackMem:= val;
               end;
            end;

            if (PriName = '.DEFINE') and (not PriFnd) then begin
               PriFnd:= true;
	             s:= GetName;
               if not IsErr then begin
                  if s <> '' then begin
                     AddDefinedName(s);
                  end else begin
                     ErrorMsg('Name expected [.DEFINE]!');
                  end;
               end;
            end;

	          if (PriName = 'DB') and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tBYTE);
            end;
            if (PriName = 'DW') and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tWORD);
            end;
            if (PriName = 'DD') and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tDWORD);
            end;
            if (PriName = 'DQ') and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tQWORD);
            end;
            if ((PriName = 'DF') or (PriName = 'DP')) and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tFARPTR);
            end;
            if (PriName = 'DT') and (not PriFnd) then begin
               PriFnd:= true;
               DefineMemData('', tTENBYTE);
            end;

            if (PriName = 'ARRAY') and (not PriFnd) then begin
               PriFnd:= true;
               DefineArray('');
            end;

            if (PriName = '.GET') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'PARAMETER' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           val:= GetNumber(tAUTO);
                           if not IsErr then begin
                              if Source[CurLevel].SrcType = srcMACRO then begin
                                 if val < Source[CurLevel].ParamValCnt then begin
                                    ss:= ParamValList[Source[CurLevel].ParamValOfs + val];
                                    pOfs:= MemLength;
                                    pCnt:= Length(ss);
                                    SetEquStr(n, pOfs, pCnt, false);
                                    if pCnt > 0 then begin
                                       i:= 0;
                                       while i < pCnt do begin
                                          AddMemValue(ord(ss[i+1]));
                                          Inc(i);
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Invalid Parameter Index [' + IntToStr(val) + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Can only get Parameter from [MACRO/OBJECT]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Source String Equate expected [.GET PARAMETER]!');
                        end;
                     end;
                  end;

                  if s = 'PARAMCOUNT' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if Source[CurLevel].SrcType = srcMACRO then begin
                              val:= Source[CurLevel].ParamValCnt;
                              SetEquInt(n, val);
                           end else begin
                              ErrorMsg('Can only get Parameter Count from [MACRO/OBJECT]!');
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'ARRAYCOUNT' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 id:= GetIdentIndex(ss);
                                 if id <> NotFound then begin
                                    if IdentList[id].IdentType = itMEM then begin
                                       val:= 1;
                                       if IdentList[id].MemType = tWORD then val:= 2;
                                       if IdentList[id].MemType = tDWORD then val:= 4;
                                       if IdentList[id].MemType = tQWORD then val:= 8;
                                       if IdentList[id].MemType = tFARPTR then val:= 6;
                                       if IdentList[id].MemType = tTENBYTE then val:= 10;
                                       val:= IdentList[id].Len div val;
                                       SetEquInt(n, val);
                                    end else begin
                                       ErrorMsg('Identifier is not an Array [' + ss + ']!');
                                    end;
                                 end else begin
                                    ErrorMsg('Identifier not found [' + ss + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Array Name expected [.GET]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'MAINSOURCE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           pOfs:= MemLength;
                           pCnt:= Length(AsmFile);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(AsmFile[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'INITDATASOURCE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           pOfs:= MemLength;
                           pCnt:= Length(IDataFile);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(IDataFile[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'ENDDATASOURCE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           pOfs:= MemLength;
                           pCnt:= Length(EDataFile);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(EDataFile[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'OUTPUTFILE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           pOfs:= MemLength;
                           pCnt:= Length(OutFile);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(OutFile[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'CURRENTFILE' then begin
                     giFnd:= true;
                     n:= GetName;
                     s:= Source[FileLevel].FileName;
                     if not IsErr then begin
                        if n <> '' then begin
                           pOfs:= MemLength;
                           pCnt:= Length(s);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(s[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'EXTRACTFILEDIR' then begin
                     giFnd:= true;
                     n:= GetName;
                     s:= GrabString;
                     if not IsErr then begin
                        if n <> '' then begin
                           s:= ExtractFileDir(s);
                           pOfs:= MemLength;
                           pCnt:= Length(s);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(s[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'EXTRACTFILEEXT' then begin
                     giFnd:= true;
                     n:= GetName;
                     s:= GrabString;
                     if not IsErr then begin
                        if n <> '' then begin
                           s:= ExtractFileExt(s);
                           pOfs:= MemLength;
                           pCnt:= Length(s);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(s[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'EXTRACTFILENAME' then begin
                     giFnd:= true;
                     n:= GetName;
                     s:= GrabString;
                     if not IsErr then begin
                        if n <> '' then begin
                           s:= ExtractFileName(s);
                           pOfs:= MemLength;
                           pCnt:= Length(s);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(s[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'EXTRACTFILEPATH' then begin
                     giFnd:= true;
                     n:= GetName;
                     s:= GrabString;
                     if not IsErr then begin
                        if n <> '' then begin
                           s:= ExtractFilePath(s);
                           pOfs:= MemLength;
                           pCnt:= Length(s);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(s[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'MAJORVERSION' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, appVerMajor);
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'MINORVERSION' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, appVerMinor);
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'RELEASE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, appVerRelease);
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'BUILD' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, appBuild);
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'DATE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= DateToStr(Date);
                           pOfs:= MemLength;
                           pCnt:= Length(ss);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(ss[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if s = 'TIME' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= TimeToStr(Time);
                           pOfs:= MemLength;
                           pCnt:= Length(ss);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(ss[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GET]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.GET]!')
                     else
                      ErrorMsg('Item expected [.GET]!');
                  end;
               end;
            end;

            if (PriName = '.GEN') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'VALUE' then begin
                     giFnd:= true;
                     typ:= tNONE;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if n = 'BYTE' then typ:= tBYTE;
                           if n = 'WORD' then typ:= tWORD;
                           if n = 'DWORD' then typ:= tDWORD;
                           if n = 'QWORD' then typ:= tQWORD;
                           if n = 'FARPTR' then typ:= tFARPTR;
                           if n = 'TENBYTE' then typ:= tTENBYTE;
                           if typ <> tNONE then begin
                              val:= GetNumber(typ);
                              if not IsErr then begin
                                 AddCodeListValue(typ, val);
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'TEXT' then begin
                     giFnd:= true;
                     ss:= GrabString;
                     if not IsErr then begin
                        if Length(ss) > 0 then begin
                           i:= 1;
                           while i <= Length(ss) do begin
                              b:= ord(ss[i]);
                              AddCodeListValue(tBYTE, b);
                              Inc(i);
                           end;
                        end;
                     end;
                  end;

                  if s = 'RELADDRESS' then begin
                     giFnd:= true;
                     typ:= tNONE;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if n = 'BYTE' then typ:= tBYTE;
                           if n = 'WORD' then typ:= tWORD;
                           if n = 'DWORD' then typ:= tDWORD;
                           if n = 'QWORD' then typ:= tQWORD;
                           if n = 'FARPTR' then typ:= tFARPTR;
                           if n = 'TENBYTE' then typ:= tTENBYTE;
                           if typ <> tNONE then begin
                              ss:= GetName;
                              if not IsErr then begin
                                 if ss <> '' then begin
                                    val:= GetCalc;
                                    if not IsErr then AddCodeListREL(typ, ss, RetOpr, val);
                                 end else begin
                                    ErrorMsg('Name expected [.GEN]!');
                                 end;
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'MEMADDRESS' then begin
                     giFnd:= true;
                     typ:= tNONE;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if n = 'BYTE' then typ:= tBYTE;
                           if n = 'WORD' then typ:= tWORD;
                           if n = 'DWORD' then typ:= tDWORD;
                           if n = 'QWORD' then typ:= tQWORD;
                           if n = 'FARPTR' then typ:= tFARPTR;
                           if n = 'TENBYTE' then typ:= tTENBYTE;
                           if typ <> tNONE then begin
                              ss:= GetName;
                              if not IsErr then begin
                                 if ss <> '' then begin
                                    val:= GetCalc;
                                    if not IsErr then AddCodeListMEM(typ, ss, RetOpr, val);
                                 end else begin
                                    ErrorMsg('Name expected [.GEN]!');
                                 end;
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'SEGOFFSET' then begin
                     giFnd:= true;
                     typ:= tNONE;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if n = 'BYTE' then typ:= tBYTE;
                           if n = 'WORD' then typ:= tWORD;
                           if n = 'DWORD' then typ:= tDWORD;
                           if n = 'QWORD' then typ:= tQWORD;
                           if n = 'FARPTR' then typ:= tFARPTR;
                           if n = 'TENBYTE' then typ:= tTENBYTE;
                           if typ <> tNONE then begin
                              ss:= GetName;
                              if not IsErr then begin
                                 if ss <> '' then begin
                                    sI:= NotFound;
                                    if SegCount > 0 then begin
                                       i:= 0;
                                       while i < SegCount do begin
                                          if Segment[i].IdentName = ss then begin
                                             sI:= i;
                                             break;
                                          end;
                                          Inc(i);
                                       end;
                                    end;
                                    if sI <> NotFound then begin
                                       val:= GetCalc;
                                       if not IsErr then AddCodeListSEGOFFSET(typ, sI, RetOpr, val);
                                    end else begin
                                       ErrorMsg('Segment not found [' + ss + '] [.GEN]!');
                                    end;
                                 end else begin
                                    ErrorMsg('Name expected [.GEN]!');
                                 end;
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'SEGSIZE' then begin
                     giFnd:= true;
                     typ:= tNONE;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if n = 'BYTE' then typ:= tBYTE;
                           if n = 'WORD' then typ:= tWORD;
                           if n = 'DWORD' then typ:= tDWORD;
                           if n = 'QWORD' then typ:= tQWORD;
                           if n = 'FARPTR' then typ:= tFARPTR;
                           if n = 'TENBYTE' then typ:= tTENBYTE;
                           if typ <> tNONE then begin
                              ss:= GetName;
                              if not IsErr then begin
                                 if ss <> '' then begin
                                    sI:= NotFound;
                                    if SegCount > 0 then begin
                                       i:= 0;
                                       while i < SegCount do begin
                                          if Segment[i].IdentName = ss then begin
                                             sI:= i;
                                             break;
                                          end;
                                          Inc(i);
                                       end;
                                    end;
                                    if sI <> NotFound then begin
                                       val:= GetCalc;
                                       if not IsErr then AddCodeListSEGSIZE(typ, sI, RetOpr, val);
                                    end else begin
                                       ErrorMsg('Segment not found [' + ss + '] [.GEN]!');
                                    end;
                                 end else begin
                                    ErrorMsg('Name expected [.GEN]!');
                                 end;
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.GEN]!');
                        end;
                     end;
                  end;



                  if s = 'GETRELOCITEM' then begin
                     giFnd:= true;
                     val:= GetNumber(tDWORD);
                     if not IsErr then begin
                        if val < RelocCount then begin
                           n:= GetName;
                           if not IsErr then begin
                              if n <> '' then begin
                                 SetEquInt(n, RelocList[val]);
                              end else begin
                                 ErrorMsg('Target Equate expected [.GEN]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Item Index out of range [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'GETRELOCLEN' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, RelocCount);
                        end else begin
                           ErrorMsg('Target Equate expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'GETPROCESSORVALUE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           if CurProcessor <> NotFound then begin
                              SetEquInt(n, ProcessorList[CurProcessor].Value);
                           end else begin
                              ErrorMsg('No Processor selected [.GEN]!');
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'GETREGCODE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 sI:= GetRegIndex(n);
                                 if sI <> NotFound then begin
                                    SetEquInt(ss, Regs[sI].Code);
                                 end else begin
                                    ErrorMsg('Register not found [' + n + '] [.GEN]!');
                                 end;
                              end else begin
                                 ErrorMsg('Target Equate expected [.GEN]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Register Name expected [.GEN]!');
                        end;
                     end;
                  end;

                  if s = 'GETREGVALUE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 sI:= GetRegIndex(n);
                                 if sI <> NotFound then begin
                                    SetEquInt(ss, Regs[sI].SegValue);
                                 end else begin
                                    ErrorMsg('Register not found [' + n + '] [.GEN]!');
                                 end;
                              end else begin
                                 ErrorMsg('Target Equate expected [.GEN]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Register Name expected [.GEN]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.GEN]!')
                     else
                      ErrorMsg('Item expected [.GEN]!');
                  end;
               end;
            end;

            if (PriName = '.PARSER') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'SET' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           id:= GetIdentIndex(ss);
                           if id <> NotFound then begin
                              if IdentList[id].IdentType = itEQU then begin
                                 if IdentList[id].EquMode = equModeSTR then begin
                                    SetParseString(IdentList[id].SValue, IdentList[id].StrLen);
                                 end else begin
                                    ErrorMsg('[' + ss + '] is not a String Equate!');
                                 end;
                              end else begin
                                 ErrorMsg('Identifier is not an Equate [' + ss + ']!');
                              end;
                           end else begin
                              ErrorMsg('[' + ss + '] not found!');
                           end;
                        end else begin
                           ErrorMsg('Equate expected!');
                        end;
                     end;
                  end;

                  if s = 'GETITEM' then begin
                     giFnd:= true;
                     aName:= GetName;
                     if not IsErr then begin
                        if aName <> '' then begin
                           val:= GetNumber(tDWORD);
                           if not IsErr then begin
                              if GetDynArrayType(aName) = tpString then begin
                                 itemDat:= GetDynArrayItem(aName, val);
                                 if not IsErr then begin
                                    SetLength(ParseObj.StrData, itemDat.StrLen+1);
                                    if itemDat.StrLen > 0 then begin
                                       i:= 0;
                                       while i < itemDat.StrLen do begin
                                          ParseObj.StrData[i]:= MemData[itemDat.StrOfs + i];
                                          Inc(i);
                                       end;
                                    end;
                                    ParseObj.StrPos:= 0;
                                    ParseObj.StrLen:= itemDat.StrLen;
                                    ParseObj.Name:= '';
                                    ParseObj.RetOpr:= oprNONE;
                                 end;
                              end else begin
                                 ErrorMsg('String Array expected [.PARSER]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETPARAM' then begin
                     giFnd:= true;
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        if Source[CurLevel].SrcType = srcMACRO then begin
                           if val < Source[CurLevel].ParamValCnt then begin
                              ss:= ParamValList[Source[CurLevel].ParamValOfs + val];
                              pCnt:= Length(ss);
                              SetLength(ParseObj.StrData, pCnt + 1);
                              if pCnt > 0 then begin
                                 i:= 0;
                                 while i < pCnt do begin
                                    ParseObj.StrData[i]:= ord(ss[i+1]);
                                    Inc(i);
                                 end;
                              end;
                              ParseObj.StrLen:= pCnt;
                              ParseObj.StrPos:= 0;
                              ParseObj.Name:= '';
                              ParseObj.RetOpr:= oprNONE;
                           end else begin
                              ErrorMsg('Invalid Parameter Index [' + IntToStr(val) + ']!');
                           end;
                        end else begin
                           ErrorMsg('Can only get Parameter from [MACRO/OBJECT]!');
                        end;
                     end;
                  end;

                  if s = 'SETPOS' then begin
                     giFnd:= true;
                     val:= GetNumber(tDWORD);
                     if not IsErr then begin
                        SetParsePos(val);
                     end;
                  end;

                  if s = 'NEXTCHAR' then begin
                     giFnd:= true;
                     ParseNextChar;
                  end;

                  if s = 'SKIPBLANK' then begin
                     giFnd:= true;
                     ParseSkipBlank;
                  end;

                  if s = 'IGNORECHAR' then begin
                     giFnd:= true;
                     val:= GetNumber(tBYTE);
                     if not IsErr then ParseIgnoreChar(val);
                  end;

                  if s = 'GETNAME' then begin
                     giFnd:= true;
                     ParseGetName;
                  end;

                  if s = 'GETNUMBER' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           t:= tNONE;
                           if n = 'BYTE' then t:= tBYTE;
                           if n = 'WORD' then t:= tWORD;
                           if n = 'DWORD' then t:= tDWORD;
                           if n = 'QWORD' then t:= tQWORD;
                           if n = 'FARPTR' then t:= tFARPTR;
                           if n = 'TENBYTE' then t:= tTENBYTE;
                           if t <> tNONE then begin
                              ss:= GetName;
                              if not IsErr then begin
                                 if ss <> '' then begin
                                    SetEquInt(ss, ParseGetNumber(t));
                                 end else begin
                                    ErrorMsg('Target Equate expected [.PARSER]!');
                                 end;
                              end;
                           end else begin
                              ErrorMsg('Invalid Type [' + n + '] [.PARSER]!');
                           end;
                        end else begin
                           ErrorMsg('Type expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETCALC' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, ParseGetCalc);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETOPR' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, ParseObj.RetOpr);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETSTRING' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           n:= ParseGetString;
                           pOfs:= MemLength;
                           pCnt:= length(n);
                           SetEquStr(ss, pOfs, pCnt, false);
                           i:= 0;
                           while i < pCnt do begin
                              AddMemValue(ord(n[i+1]));
                              Inc(i);
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETBASICSTRING' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           n:= ParseGetSimpleString;
                           pOfs:= MemLength;
                           pCnt:= length(n);
                           SetEquStr(ss, pOfs, pCnt, false);
                           i:= 0;
                           while i < pCnt do begin
                              AddMemValue(ord(n[i+1]));
                              Inc(i);
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETNUM' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, ParseGetNum);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'NAMETOEQU' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           n:= ParseObj.Name;
                           pOfs:= MemLength;
                           pCnt:= length(n);
                           SetEquStr(ss, pOfs, pCnt, false);
                           i:= 0;
                           while i < pCnt do begin
                              AddMemValue(ord(n[i+1]));
                              Inc(i);
                           end;
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'ISENDLINE' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           if ParseIsEndLine then
                            SetEquInt(ss, TrueVal)
                           else
                            SetEquInt(ss, FalseVal);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETPOS' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, ParseObj.StrPos);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETLEN' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, ParseObj.StrLen);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETCHAR' then begin
                     giFnd:= true;
                     ss:= GetName;
                     if not IsErr then begin
                        if ss <> '' then begin
                           SetEquInt(ss, GetParseChar);
                        end else begin
                           ErrorMsg('Target Equate expected [.PARSER]!');
                        end;
                     end;
                  end;

                  if s = 'GETCHARAT' then begin
                     giFnd:= true;
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        ss:= GetName;
                        if not IsErr then begin
                           if ss <> '' then begin
                              SetEquInt(ss, GetParseCharAt(val));
                           end else begin
                              ErrorMsg('Target Equate expected [.PARSER]!');
                           end;
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.PARSER]!')
                     else
                      ErrorMsg('Item expected [.PARSER]!');
                  end;
               end;
            end;

            {if (PriName = '.IDENT') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = '..' then begin
                     giFnd:= true;
                      
                  end;



                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.IDENT]!')
                     else
                      ErrorMsg('Item expected [.IDENT]!');
                  end;
               end;
            end;  }

            if (PriName = '.FILE') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'NEW' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GrabString;
                           if not IsErr then begin
                              AddFileObj(n, ss, fmNone);
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'OPEN' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GrabString;
                           if not IsErr then begin
                              m:= GetName;
                              if not IsErr then begin
                                 if m <> '' then begin
                                    if m = 'WRITE' then begin
                                       OpenFileObj(n, ss, fmOpenWrite);
                                    end else begin
                                       if m = 'READ' then begin
                                          OpenFileObj(n, ss, fmOpenRead);
                                       end else begin
                                          ErrorMsg('Invalid File mode [' + m + ']!');
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('[WRITE or READ] expected [.FILE]!');
                                 end;
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'CLOSE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           CloseFileObj(n);
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'SEEK' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           val:= GetNumber(tDWORD);
                           if not IsErr then begin
                              FileObjSeek(n, val);
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'GETPOS' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetFileObjPos(n));
                              end else begin
                                 ErrorMsg('Target Equate Expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'GETLEN' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetFileObjLen(n));
                              end else begin
                                 ErrorMsg('Target Equate Expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'GETMODE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetFileMode(n));
                              end else begin
                                 ErrorMsg('Target Equate Expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'GETERROR' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetFileError(n));
                              end else begin
                                 ErrorMsg('Target Equate Expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'ISEOF' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 if IsFileObjEOF(n) then
                                  SetEquInt(ss, TrueVal)
                                 else
                                  SetEquInt(ss, FalseVal);
                              end else begin
                                 ErrorMsg('Target Equate Expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'WRITE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 b:= 255;
                                 if ss = 'VALUE' then b:= 0;
                                 if ss = 'STRING' then b:= 1;
                                 if ss = 'BLOCK' then b:= 2;
                                 if ss = 'STRINGLN' then b:= 3;
                                 if b <> 255 then begin
                                    if b = 0 then begin
                                       ss:= GetName;
                                       if not IsErr then begin
                                          if ss <> '' then begin
                                             typ:= tNone;
                                             if ss = 'BYTE' then typ:= tBYTE;
                                             if ss = 'WORD' then typ:= tWORD;
                                             if ss = 'DWORD' then typ:= tDWORD;
                                             if ss = 'QWORD' then typ:= tQWORD;
                                             if ss = 'FARPTR' then typ:= tFARPTR;
                                             if ss = 'TENBYTE' then typ:= tTENBYTE;
                                             if typ <> tNone then begin
                                                val:= GetNumber(typ);
                                                if not IsErr then begin
                                                   WriteFileObjValue(n, typ, val);
                                                end;
                                             end else begin
                                                ErrorMsg('Invalid Type [' + ss + '] [.FILE]!');
                                             end;
                                          end else begin
                                             ErrorMsg('Type expected [.FILE]!');
                                          end;
                                       end;
                                    end;
                                    if b = 1 then begin
                                       aName:= GrabString;
                                       if not IsErr then begin
                                          WriteFileObjString(n, aName, false);
                                       end;
                                    end;
                                    if b = 2 then begin
                                       ss:= GetName;
                                       if not IsErr then begin
                                          if ss <> '' then begin
                                             typ:= tNone;
                                             if ss = 'BYTE' then typ:= tBYTE;
                                             if ss = 'WORD' then typ:= tWORD;
                                             if ss = 'DWORD' then typ:= tDWORD;
                                             if ss = 'QWORD' then typ:= tQWORD;
                                             if ss = 'FARPTR' then typ:= tFARPTR;
                                             if ss = 'TENBYTE' then typ:= tTENBYTE;
                                             if typ <> tNone then begin
                                                aName:= GetName;
                                                if not IsErr then begin
                                                   if aName <> '' then begin
                                                      val:= GetNumber(tDWORD);
                                                      if not IsErr then begin
                                                         cnt:= GetNumber(tDWORD);
                                                         if not IsErr then begin
                                                            WriteFileObjBlock(n, aName, typ, val, cnt);
                                                         end;
                                                      end;
                                                   end else begin
                                                      ErrorMsg('Dynamic Array Name expected [.FILE]!');
                                                   end;
                                                end;
                                             end else begin
                                                ErrorMsg('Invalid Type [' + ss + '] [.FILE]!');
                                             end;
                                          end else begin
                                             ErrorMsg('Type expected [.FILE]!');
                                          end;
                                       end;
                                    end;
                                    if b = 3 then begin
                                       aName:= GrabString;
                                       if not IsErr then begin
                                          WriteFileObjString(n, aName, true);
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Invalid Write method [' + ss + '] [.FILE]!');
                                 end;
                              end else begin
                                 ErrorMsg('[VALUE / STRING / STRINGLN / BLOCK] expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if s = 'READ' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 b:= 255;
                                 if ss = 'VALUE' then b:= 0;
                                 if ss = 'STRING' then b:= 1;
                                 if ss = 'BLOCK' then b:= 2;
                                 if b <> 255 then begin
                                    if b = 0 then begin
                                       ss:= GetName;
                                       if not IsErr then begin
                                          if ss <> '' then begin
                                             typ:= tNone;
                                             if ss = 'BYTE' then typ:= tBYTE;
                                             if ss = 'WORD' then typ:= tWORD;
                                             if ss = 'DWORD' then typ:= tDWORD;
                                             if ss = 'QWORD' then typ:= tQWORD;
                                             if ss = 'FARPTR' then typ:= tFARPTR;
                                             if ss = 'TENBYTE' then typ:= tTENBYTE;
                                             if typ <> tNone then begin
                                                tName:= GetName;
                                                if not IsErr then begin
                                                   if tName <> '' then begin
                                                      SetEquInt(tName, ReadFileObjValue(n, typ));
                                                   end else begin
                                                      ErrorMsg('Target Equate expected [.FILE]!');
                                                   end;
                                                end;
                                             end else begin
                                                ErrorMsg('Invalid Type [' + ss + '] [.FILE]!');
                                             end;
                                          end else begin
                                             ErrorMsg('Type expected [.FILE]!');
                                          end;
                                       end;
                                    end;
                                    if b = 1 then begin
                                       tName:= GetName;
                                       if not IsErr then begin
                                          if tName <> '' then begin
                                             ss:= ReadFileObjString(n);
                                             if not IsErr then begin
                                                pOfs:= MemLength;
                                                pCnt:= Length(ss);
                                                SetEquStr(tName, pOfs, pCnt, false);
                                                if pCnt > 0 then begin
                                                   i:= 0;
                                                   while i < pCnt do begin
                                                      AddMemValue(ord(ss[i+1]));
                                                      Inc(i);
                                                   end;
                                                end;
                                             end;
                                          end else begin
                                             ErrorMsg('Target Equate expected [.FILE]!');
                                          end;
                                       end;
                                    end;
                                    if b = 2 then begin
                                       ss:= GetName;
                                       if not IsErr then begin
                                          if ss <> '' then begin
                                             typ:= tNone;
                                             if ss = 'BYTE' then typ:= tBYTE;
                                             if ss = 'WORD' then typ:= tWORD;
                                             if ss = 'DWORD' then typ:= tDWORD;
                                             if ss = 'QWORD' then typ:= tQWORD;
                                             if ss = 'FARPTR' then typ:= tFARPTR;
                                             if ss = 'TENBYTE' then typ:= tTENBYTE;
                                             if typ <> tNone then begin
                                                aName:= GetName;
                                                if not IsErr then begin
                                                   if aName <> '' then begin
                                                      val:= GetNumber(tDWORD);
                                                      if not IsErr then begin
                                                         cnt:= GetNumber(tDWORD);
                                                         if not IsErr then begin
                                                            ReadFileObjBlock(n, aName, typ, val, cnt);
                                                         end;
                                                      end;
                                                   end else begin
                                                      ErrorMsg('Dynamic Array Name expected [.FILE]!');
                                                   end;
                                                end;
                                             end else begin
                                                ErrorMsg('Invalid Type [' + ss + '] [.FILE]!');
                                             end;
                                          end else begin
                                             ErrorMsg('Type expected [.FILE]!');
                                          end;
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Invalid Read method [' + ss + '] [.FILE]!');
                                 end;
                              end else begin
                                 ErrorMsg('[VALUE / STRING / BLOCK] expected [.FILE]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('File Object Name expected [.FILE]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.FILE]!')
                     else
                      ErrorMsg('Item expected [.FILE]!');
                  end;
               end;
            end;

            if (PriName = '.DYNARRAY') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'NEW' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 b:= 255;
                                 if ss = 'VALUE' then b:= tpValue;
                                 if ss = 'STRING' then b:= tpString;
                                 if ss = 'NAME' then b:= tpName;
                                 if b <> 255 then begin
                                    val:= GetNumber(tDWORD);
                                    if not IsErr then begin
                                       AddDynArray(n, b, val);
                                    end;
                                 end else begin
                                    ErrorMsg('Invalid Dynamic Array Type [' + ss + ']!');
                                 end;
                              end else begin
                                 ErrorMsg('Dynamic Array Type expected [.DYNARRAY]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'SETLENGTH' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           val:= GetNumber(tAUTO);
                           if not IsErr then begin
                              SetDynArrayLength(n, val);
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'GETLENGTH' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetDynArrayLength(n));
                              end else begin
                                 ErrorMsg('Target Equate expected [.DYNARRAY]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'GETTYPE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GetName;
                           if not IsErr then begin
                              if ss <> '' then begin
                                 SetEquInt(ss, GetDynArrayType(n));
                              end else begin
                                 ErrorMsg('Target Equate expected [.DYNARRAY]!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'SETITEM' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           id:= GetDynArrayIndex(n);
                           typ:= GetDynArrayType(n);
                           if not IsErr then begin
                              indx:= GetNumber(tDWORD);
                              if not IsErr then begin
                                 if indx <= GetDynArrayLength(n) then begin
                                    if typ = tpValue then begin
                                       val:= GetNumber(tAUTO);
                                       if not IsErr then begin
                                          DynArrayList[id].Items[indx].Value:= val;
                                          DynArrayList[id].Items[indx].StrOfs:= 0;
                                          DynArrayList[id].Items[indx].StrLen:= 0;
                                          DynArrayList[id].Items[indx].Name:= '';
                                       end;
                                    end else begin
                                       if typ = tpString then begin
                                          ss:= GrabString;
                                          if not IsErr then begin
                                             MemReplacePos:= -1;
                                             pOfs:= MemLength;
                                             pCnt:= length(ss);
                                             if pCnt > 0 then begin
                                                i:= 0;
                                                while i < pCnt do begin
                                                   b:= ord(ss[i+1]);
                                                   AddMemValue(b);
                                                   Inc(i);
                                                end;
                                             end;
                                             DynArrayList[id].Items[indx].Value:= 0;
                                             DynArrayList[id].Items[indx].StrOfs:= pOfs;
                                             DynArrayList[id].Items[indx].StrLen:= pCnt;
                                             DynArrayList[id].Items[indx].Name:= '';
                                          end;
                                       end else begin
                                          ss:= GetName;
                                          if not IsErr then begin
                                             if ss <> '' then begin
                                                DynArrayList[id].Items[indx].Value:= 0;
                                                DynArrayList[id].Items[indx].StrOfs:= 0;
                                                DynArrayList[id].Items[indx].StrLen:= 0;
                                                DynArrayList[id].Items[indx].Name:= ss;
                                             end else begin
                                                ErrorMsg('Name expected [.DYNARRAY]');
                                             end;
                                          end;
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('Dynamic Array Index out of limit [' + n + ']!');
                                 end;
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'GETITEM' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           indx:= GetNumber(tDWORD);
                           if not IsErr then begin
                              if indx <= GetDynArrayLength(n) then begin
                                 ss:= GetName;
                                 if not IsErr then begin
                                    if ss <> '' then begin
                                       i:= GetDynArrayIndex(n);
                                       if GetDynArrayType(n) = tpValue then begin
                                           SetEquInt(ss, DynArrayList[i].Items[indx].Value);
                                       end else begin
                                          if GetDynArrayType(n) = tpString then begin
                                             SetEquStr(ss, DynArrayList[i].Items[indx].StrOfs, DynArrayList[i].Items[indx].StrLen, true);
                                          end else begin
                                             ErrorMsg('[GETITEM] is not allowed in Name-Array, use [GETNAME] instead [.DYNARRAY]!');
                                          end;
                                       end;
                                    end else begin
                                       ErrorMsg('Target Equate expected [.DYNARRAY]!');
                                    end;
                                 end;
                              end else begin
                                 ErrorMsg('Dynamic Array Index out of limit [' + n + ']!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if s = 'GETNAME' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           indx:= GetNumber(tDWORD);
                           if not IsErr then begin
                              if indx <= GetDynArrayLength(n) then begin
                                 i:= GetDynArrayIndex(n);
                                 if GetDynArrayType(n) = tpName then begin
                                    DynArrayItemName:= DynArrayList[i].Items[indx].Name;
                                 end else begin
                                    ErrorMsg('[GETNAME] is not allowed in Value-/String-Array, use [GETITEM] instead [.DYNARRAY]!');
                                 end;
                              end else begin
                                 ErrorMsg('Dynamic Array Index out of limit [' + n + ']!');
                              end;
                           end;
                        end else begin
                           ErrorMsg('Dynamic Array Name expected [.DYNARRAY]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.DYNARRAY]!')
                     else
                      ErrorMsg('Item expected [.DYNARRAY]!');
                  end;
               end;
            end;

            if (PriName = '.DISK') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'FILEEXISTS' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GrabString;
                           if not IsErr then begin
                              if FileExists(ss) then
                               SetEquInt(n, TrueVal)
                              else
                               SetEquInt(n, FalseVal);
                           end;
                        end else begin
                           ErrorMsg('Equate expected [.DISK]!');
                        end;
                     end;
                  end;

                  if s = 'GETDIR' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           GetDir(0, ss);
                           pOfs:= MemLength;
                           pCnt:= Length(ss);
                           SetEquStr(n, pOfs, pCnt, false);
                           if pCnt > 0 then begin
                              i:= 0;
                              while i < pCnt do begin
                                 AddMemValue(ord(ss[i+1]));
                                 Inc(i);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Equate expected [.DISK]!');
                        end;
                     end;
                  end;

                  if s = 'GETDISKDIR' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           val:= GetNumber(tBYTE);
                           if val > (ord('Z')-ord('A'))+1 then
                            ErrorMsg('Invalid Drive ID [.DISK]!');
                           if not IsErr then begin
                              GetDir(val, ss);
                              pOfs:= MemLength;
                              pCnt:= Length(ss);
                              SetEquStr(n, pOfs, pCnt, false);
                              if pCnt > 0 then begin
                                 i:= 0;
                                 while i < pCnt do begin
                                    AddMemValue(ord(ss[i+1]));
                                    Inc(i);
                                 end;
                              end;
                           end;
                        end else begin
                           ErrorMsg('Equate expected [.DISK]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.DISK]!')
                     else
                      ErrorMsg('Item expected [.DISK]!');
                  end;
               end;
            end;

            if (PriName = '.STRING') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'SET' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ss:= GrabString;
                           if not IsErr then begin
                              pOfs:= MemLength;
                              pCnt:= Length(ss);
                              SetEquStr(n, pOfs, pCnt, false);
                              if pCnt > 0 then begin
                                 i:= 0;
                                 while i < pCnt do begin
                                    AddMemValue(ord(ss[i+1]));
                                    Inc(i);
                                 end;
                              end;
                           end;
                        end else begin
                           ErrorMsg('String Equate expected [.STRING]!');
                        end;
                     end;
                  end;

                  if s = 'TOUPPER' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           id:= GetIdentIndex(n);
                           if id <> NotFound then begin
                              if IdentList[id].IdentType = itEQU then begin
                                 if IdentList[id].EquMode = equModeSTR then begin
                                    pOfs:= IdentList[id].SValue;
                                    pCnt:= IdentList[id].StrLen;
                                    if pCnt > 0 then begin
                                       i:= 0;
                                       while i < pCnt do begin
                                          c:= chr(MemData[pOfs + i]);
                                          c:= UpCase(c);
                                          MemData[pOfs + i]:= ord(c);
                                          Inc(i);
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('[' + n + '] is not a String Equate [.STRING]!');
                                 end;
                              end else begin
                                 ErrorMsg('[' + n + '] is not a String Equate [.STRING]!');
                              end;
                           end else begin
                              ErrorMsg('[' + n + '] not found!');
                           end;
                        end else begin
                           ErrorMsg('String Equate expected [.STRING]!');
                        end;
                     end;
                  end;

                  if s = 'TOLOWER' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           id:= GetIdentIndex(n);
                           if id <> NotFound then begin
                              if IdentList[id].IdentType = itEQU then begin
                                 if IdentList[id].EquMode = equModeSTR then begin
                                    pOfs:= IdentList[id].SValue;
                                    pCnt:= IdentList[id].StrLen;
                                    if pCnt > 0 then begin
                                       i:= 0;
                                       while i < pCnt do begin
                                          c:= chr(MemData[pOfs + i]);
                                          ss:= LowerCase('' + c);
                                          c:= ss[1];
                                          MemData[pOfs + i]:= ord(c);
                                          Inc(i);
                                       end;
                                    end;
                                 end else begin
                                    ErrorMsg('[' + n + '] is not a String Equate [.STRING]!');
                                 end;
                              end else begin
                                 ErrorMsg('[' + n + '] is not a String Equate [.STRING]!');
                              end;
                           end else begin
                              ErrorMsg('[' + n + '] not found!');
                           end;
                        end else begin
                           ErrorMsg('String Equate expected [.STRING]!');
                        end;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.STRING]!')
                     else
                      ErrorMsg('Item expected [.STRING]!');
                  end;
               end;
            end;

            if (PriName = '.MODRM') and (not PriFnd) then begin
               PriFnd:= true;
               s:= GetName;
               giFnd:= false;
               if not IsErr then begin
                  if s = 'RESET' then begin
                     giFnd:= true;
                     ResetModRMInput;
                  end;

                  if s = 'GETRETURNVALUE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           SetEquInt(n, ModRMInput.ReturnValue);
                        end else begin
                           ErrorMsg('Target Equate expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'GENSTANDARD' then begin
                     giFnd:= true;
                     ModRMInput.ReturnValue:= GetModRM(ModRMInput.ModTable,
                      ModRMInput.MemBits,
                      ModRMInput.BracketCode, ModRMInput.Reg1, ModRMInput.Reg2,
                      ModRMInput.Reg3, ModRMInput.Reg);
                  end;

                  if s = 'GENDIGIT' then begin
                     giFnd:= true;
                     ModRMInput.ReturnValue:= GetModRMDigit(ModRMInput.ModTable,
                      ModRMInput.MemBits,
                      ModRMInput.BracketCode, ModRMInput.Reg1, ModRMInput.Reg2,
                      ModRMInput.Reg3, ModRMInput.Digit);
                  end;

                  if s = 'GENSPECIAL' then begin
                     giFnd:= true;
                     ModRMInput.ReturnValue:= GetModRMSpecial(ModRMInput.ModTable,
                      ModRMInput.Value, ModRMInput.Reg);
                  end;

                  if s = 'SETTABLE' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ModRMInput.ModTable:= n;
                        end else begin
                           ErrorMsg('Table Name expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'SETBITS' then begin
                     giFnd:= true;
                     val:= GetNumber(tAUTO);
                     if not IsErr then ModRMInput.MemBits:= val;
                  end;

                  if s = 'SETREG' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ModRMInput.Reg:= n;
                        end else begin
                           ErrorMsg('Register/Memory Name expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'SETREG1' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ModRMInput.Reg1:= n;
                        end else begin
                           ErrorMsg('Register/Memory Name expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'SETREG2' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ModRMInput.Reg2:= n;
                        end else begin
                           ErrorMsg('Register/Memory Name expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'SETREG3' then begin
                     giFnd:= true;
                     n:= GetName;
                     if not IsErr then begin
                        if n <> '' then begin
                           ModRMInput.Reg3:= n;
                        end else begin
                           ErrorMsg('Register/Memory Name expected [.MODRM]!');
                        end;
                     end;
                  end;

                  if s = 'CLEARTABLE' then begin
                     giFnd:= true;
                     ModRMInput.ModTable:= '';
                  end;

                  if s = 'CLEARREG' then begin
                     giFnd:= true;
                     ModRMInput.Reg:= '';
                  end;

                  if s = 'CLEARREG1' then begin
                     giFnd:= true;
                     ModRMInput.Reg1:= '';
                  end;

                  if s = 'CLEARREG2' then begin
                     giFnd:= true;
                     ModRMInput.Reg2:= '';
                  end;

                  if s = 'CLEARREG3' then begin
                     giFnd:= true;
                     ModRMInput.Reg3:= '';
                  end;

                  if s = 'SETBRACKETCODE' then begin
                     giFnd:= true;
                     val:= GetNumber(tBYTE);
                     if not IsErr then begin
                        ModRMInput.BracketCode:= val;
                     end;
                  end;

                  if s = 'SETDIGIT' then begin
                     giFnd:= true;
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        ModRMInput.Digit:= val;
                     end;
                  end;

                  if s = 'SETVALUE' then begin
                     giFnd:= true;
                     val:= GetNumber(tAUTO);
                     if not IsErr then begin
                        ModRMInput.Value:= val;
                     end;
                  end;

                  if not giFnd then begin
                     if s <> '' then
                      ErrorMsg('Invalid Item [' + s + '] [.MODRM]!')
                     else
                      ErrorMsg('Item expected [.MODRM]!');
                  end;
               end;
            end;

            if not PriFnd then begin
               if PriName <> '' then begin
                  executeCmd:= false;
                  id:= GetIdentIndex(PriName);
                  if id <> NotFound then begin
                     if IdentList[id].IdentType = itMACRO then begin
                        executeCmd:= true;
                        pOfs:= ParamValCount;
                        pCnt:= GrabParamValues(pOfs);
                        if not IsErr then begin
                           CheckEndLine;
                           if not IsErr then begin
                              Inc(Source[CurLevel].CurLine);
                              Inc(CurID);
                              Inc(CurLevel);
                              SetLength(Source, CurLevel + 1);
                              SetSource(CurLevel, CurID, PriName, srcMACRO);
                              Source[CurLevel].SrcOffset:= IdentList[id].MacroOffset;
                              Source[CurLevel].CurLine:= 0;
                              Source[CurLevel].NumLines:= IdentList[id].MacroLen;
                              Source[CurLevel].SrcLine:= '';
                              Source[CurLevel].Pos:= Source[CurLevel].Len + 1;
                              Source[CurLevel].ParamValOfs:= pOfs;
                              Source[CurLevel].ParamValCnt:= pCnt;
                              sOff:= Source[CurLevel].SrcOffset;
                              nLin:= Source[CurLevel].NumLines;
                              AddSourceData(CurLevel, srcMACRO, sOff, nLin, PriName);
                           end;
                        end;
                     end;
                     if IdentList[id].IdentType = itOBJECT then begin
                        executeCmd:= true;
                        mainParam:= GetName;
                        if mainParam <> '' then begin
                           pOfs:= ParamValCount;
                           AddParamValue(mainParam);
                           pCnt:= 1;
                           SkipBlank;
                           if GetSrcChar = '(' then begin
                              NextChar;
                              pCnt:= GrabObjectParamValues(pOfs);
                              SkipBlank;
                              if GetSrcChar = ')' then begin
                                 NextChar;
                              end else begin
                                 ErrorMsg('[ ) ] expected!');
                              end;
                           end else begin
                              ErrorMsg('[ ( ] expected!');
                           end;
                           if not IsErr then begin
                              CheckEndLine;
                              if not IsErr then begin
                                 Inc(Source[CurLevel].CurLine);
                                 Inc(CurID);
                                 Inc(CurLevel);
                                 SetLength(Source, CurLevel + 1);
                                 SetSource(CurLevel, CurID, PriName, srcMACRO); { = srcOBJECT??? }
                                 Source[CurLevel].SrcOffset:= IdentList[id].MacroOffset;
                                 Source[CurLevel].CurLine:= 0;
                                 Source[CurLevel].NumLines:= IdentList[id].MacroLen;
                                 Source[CurLevel].SrcLine:= '';
                                 Source[CurLevel].Pos:= Source[CurLevel].Len + 1;
                                 Source[CurLevel].ParamValOfs:= pOfs;
                                 Source[CurLevel].ParamValCnt:= pCnt;
				                         sOff:= Source[CurLevel].SrcOffset;
                                 nLin:= Source[CurLevel].NumLines;
                                 AddSourceData(CurLevel, srcOBJECT, sOff, nLin, PriName);
                              end;
                           end;
                        end else begin
                           ErrorMsg('Object name expected!');
                        end;
                     end;
                  end else begin
                     segI:= Neg;
                     i:= 0;
                     while i < SegCount do begin
                        if Segment[i].IdentName = PriName then begin
                           segI:= i;
                           break;
                        end;
                        Inc(i);
                     end;
                     if segI <> Neg then begin
                        executeCmd:= true;
                        if SecName = '' then begin
                           CurSeg:= segI;
                        end else begin
                           SecSegment:= segI;
                        end;
                     end else begin
                        procI:= Neg;
                        i:= 0;
                        while i < ProcessorCount do begin
                           if ProcessorList[i].Name = PriName then begin
                              procI:= i;
                              break;
                           end;
                           Inc(i);
                        end;
                        if procI <> Neg then begin
                           executeCmd:= true;
                           CurProcessor:= procI;
                        end else begin
                           outI:= Neg;
                           i:= 0;
                           while i < OutputCount do begin
                              if OutputList[i].IdentName = PriName then begin
                                 outI:= i;
                                 break;
                              end;
                              Inc(i);
                           end;
                           if outI <> Neg then begin
                              executeCmd:= true;
                              CurOutput:= outI;
                           end else begin
                              RunInstr(PriName);
                              if (IsErr) or (RunInstr2) then
                               executeCmd:= true;
                           end;
                        end;
                     end;
                  end;

                  if not executeCmd then begin
                     if GetSrcChar = ':' then begin
                        NextChar;
		                    val:= GetSegOffset(false);
                        AddLabel(PriName, val);
                     end else begin
                        SkipBlank;
                        if GetSrcChar = '=' then begin
                           NextChar;
                           id:= GetIdentIndex(PriName);
                           { val:= 0; }
                           if id = NotFound then begin
                              AddEquInt(PriName, 0);
                              id:= GetIdentIndex(PriName);
                           end;
                           val:= GetNumber(tAUTO);
                           if not IsErr then begin
                              if IdentList[id].IdentType = itEQU then begin
                                 IdentList[id].IValue:= val;
                                 IdentList[id].EquMode:= equModeINT;
                                 IdentList[id].EquIdentIndex:= Neg;
                                 IdentList[id].EquOpr:= oprNONE;
                                 IdentList[id].EquName:= '';
                                 IdentList[id].SValue:= 0;
                                 IdentList[id].StrLen:= 0;
                                 IdentList[id].RValue:= 0.0;
                              end else begin
                                 ErrorMsg('Identifier is not an Equate [' + PriName + ']!');
                              end;
                           end;
                        end else begin
                           SubName:= GetName;
                           if SubName <> '' then begin
                              SubFnd:= false;
                              if (SubName = 'MACRO') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 MakeMacro(PriName);
                              end;
                              if (SubName = 'OBJECT') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 MakeObject(PriName);
                              end;
                              if (SubName = 'PROC') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 if not IsErr then begin
                                    if ProcName = '' then begin
                                       ProcName:= PriName;
                                       ProcOffset:= GetSegOffset(false);
                                    end else begin
                                       ErrorMsg('Can not create a procedure inside another procedure!');
                                    end;
                                 end;
                              end;
                              if (SubName = 'SECTION') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 if not IsErr then begin
                                    if SecName = '' then begin
                                       if ProcName = '' then begin
                                          uSegName:= GetName;
                                          uSeg:= CurSeg;
                                          if uSegName <> '' then begin
                                             uSeg:= Neg;
                                             i:= 0;
                                             while i < segCount do begin
                                                if uSegName = Segment[i].IdentName then begin
                                                   uSeg:= i;
                                                   break;
                                                end;
                                                Inc(i);
                                             end;
                                             if uSeg = Neg then begin
                                                ErrorMsg('Segment not found [' + uSegName + ']!');
                                             end;
                                          end;
                                          if not IsErr then begin
                                             SecName:= PriName;
                                             SecSegment:= uSeg;
                                             SecOffset:= GetSegOffset(true);
                                             SecPos:= 0;
                                             SecORG:= 0;
                                             SecRound:= 0;
                                          end;
                                       end else begin
                                          ErrorMsg('Can not create a section inside a procedure!');
                                       end;
                                    end else begin
                                       ErrorMsg('Can not create a section inside another section!');
                                    end;
                                 end;
                              end;

			                        if (SubName = 'ENDM') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 ErrorMsg('[ENDM] without [MACRO]!');
                              end;
                              if (SubName = 'ENDO') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 ErrorMsg('[ENDO] without [OBJECT]!');
                              end;
                              if (SubName = 'ENDP') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 if ProcName <> '' then begin
                                    if PriName = ProcName then begin
                                       if not IsErr then begin
                                          AddLabel(ProcName, ProcOffset);
                                          ProcName:= '';
                                       end;
                                    end else begin
                                       ErrorMsg('[' + PriName + '] does not match procedure name!');
                                    end;
                                 end else begin
                                    ErrorMsg('[ENDP] without [PROC]!');
                                 end;
                              end;
                              if (SubName = 'ENDS') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 if SecName <> '' then begin
                                    if ProcName = '' then begin
                                       if PriName = SecName then begin
                                          if not IsErr then begin
                                             AddSection(SecName, SecOffset, SecPos);
                                             SecName:= '';
                                          end;
                                       end else begin
                                          ErrorMsg('[' + PriName + '] does not match section name!');
                                       end;
                                    end else begin
                                       ErrorMsg('[ENDP] expected, but [ENDS] found!');
                                    end;
                                 end else begin
                                    ErrorMsg('[ENDS] without [SECTION]!');
                                 end;
                              end;

			                        if (SubName = 'DB') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tBYTE);
                              end;
                              if (SubName = 'DW') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tWORD);
                              end;
                              if (SubName = 'DD') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tDWORD);
                              end;
                              if (SubName = 'DQ') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tQWORD);
                              end;
                              if ((SubName = 'DF') or (SubName = 'DP')) and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tFARPTR);
                              end;
                              if (SubName = 'DT') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineMemData(PriName, tTENBYTE);
                              end;

                              if (SubName = 'ARRAY') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineArray(PriName);
                              end;

			                        if (SubName = 'EQU') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineEqu(PriName);
                              end;
                              if (SubName = 'EQUS') and (not SubFnd) then begin
                                 SubFnd:= true;
                                 DefineEquStr(PriName);
                              end;

                              if not SubFnd then begin
                                 if GetIdentIndex(PriName) <> NotFound then begin
                                    ErrorMsg('Illegal use of identifier [' + PriName + ']!');
                                 end else begin
                                    if GetIdentIndex(SubName) <> NotFound then begin
                                       ErrorMsg('Illegal use of identifier [' + SubName + ']!');
                                    end else begin
                                       ErrorMsg('Unknown identifiers [' + PriName + '] [' + SubName + ']!');
                                    end;
                                 end;
                              end;
                           end else begin
                              if GetIdentIndex(PriName) <> NotFound then begin
                                 ErrorMsg('Illegal use of identifier [' + PriName + ']!');
                              end else begin
                                 ErrorMsg('Unknown identifier [' + PriName + ']!');
                              end;
                           end;
                        end;
                     end;
                  end else begin
                     if not IsErr then CheckEndLine;
                  end;
               end else begin
                  SkipBlank;
                  if (Source[CurLevel].Pos <= Source[CurLevel].Len)
                   and (GetSrcChar <> ';') then ErrorMsg('Syntax Error!');
               end;
               { break; }
            end;
            if not IsErr then CheckEndLine;
         end;
         {if not IsErr then CheckEndLine; }
      end else begin
         if (State) and (not IsErr) then CheckEndLine;
      end;
   end else begin
      if not IsErr then CheckEndLine;
   end;
end;



procedure CheckEndLine;
begin
   SkipBlank;
   if (Source[CurLevel].Pos <= Source[CurLevel].Len)
    and (GetSrcChar <> ';') then
     ErrorMsg('Syntax Error! [Expected end of line]!');
end;

procedure CheckStructEnds;
begin
   if MacroName <> '' then begin
      if not MacroIsObject then begin
         FatalError('[MACRO] [' + MacroName + '] without [ENDM] [' + Source[CurLevel].FileName+ ']!');
      end else begin
         FatalError('[OBJECT] [' + MacroName + '] without [ENDO] [' + Source[CurLevel].FileName+ ']!');
      end;
      MacroName:= '';
   end;
   if ProcName <> '' then begin
      FatalError('[PROC] [' + ProcName + '] without [ENDP] [' + Source[CurLevel].FileName+ ']!');
      ProcName:= '';
   end;
   if SecName <> '' then begin
      FatalError('[SECTION] [' + SecName + '] without [ENDS] [' + Source[CurLevel].FileName+ ']!');
      SecName:= '';
   end;
end;

procedure CheckParam(parName: string;
 parOfs: TDefValue; parCnt: TAsmValue);
var
 i                 : TDefValue;
begin
   i:= 0;
   while i < parCnt do begin
      if ParamList[parOfs + i] = parName then begin
         ErrorMsg('Parameter is already added [' + parName + ']!');
         break;
      end;
      Inc(i);
   end;
end;



function GrabParams(prmOfs: TDefValue; isObj: boolean): TAsmValue;
var
 prCnt             : TAsmValue;
 prOfs             : TDefValue;
 getParam          : boolean;
 needParam         : boolean;
 pName             : string;
begin
   prCnt:= 0;
   if isObj then prCnt:= 1;
   prOfs:= prmOfs;
   getParam:= true;
   needParam:= false;
   while getParam do begin
      pName:= GetName;
      if pName <> '' then begin
         CheckParam(pName, prOfs, prCnt);
         if not IsErr then begin
            AddParam(pName);
            Inc(prCnt);
            SkipBlank;
            if GetSrcChar = ',' then begin
               NextChar;
               needParam:= true;
            end else begin
               getParam:= false;
            end;
         end else begin
            getParam:= false;
         end;
      end else begin
         if needParam then begin
            ErrorMsg('Need parameter name after [ , ]!');
         end;
         getParam:= false;
      end;
      if IsErr then getParam:= false;
   end;
   GrabParams:= prCnt;
end;



procedure MakeMacro(macName: string);
var
 MName             : string;
 pOfs              : TDefValue;
 pCnt              : TAsmValue;
begin
   if Source[CurLevel].StateLevel > 0 then begin
      ErrorMsg('Can not create macros inside [.IF], [.FOR] and/or [.WHILE]!');
   end;
   if MacroName <> '' then begin
      ErrorMsg('Macros can not be created inside another macro or object!');
   end;
   if ProcName <> '' then begin
      ErrorMsg('Can not create macros inside procedures!');
   end;
   if SecName <> '' then begin
      ErrorMsg('Can not create macros inside sections!');
   end;
   if not IsErr then begin
      MName:= '';
      if macName <> '' then begin
         MName:= macName;
      end else begin
         MName:= GetName;
      end;
      if MName <> '' then begin
         pOfs:= ParamCnt;
         pCnt:= GrabParams(pOfs, false);
         if not IsErr then begin
            MacroName:= MName;
            MacroIsObject:= false;
            MacroStart:= Source[CurLevel].SrcOffset + Source[CurLevel].CurLine;
            MacroLength:= 0;
            MacroParamOffset:= pOfs;
            MacroParamCount:= pCnt;
         end;
      end else begin
         ErrorMsg('Macro name expected!');
      end;
   end;
end;

procedure MakeObject(objName: string);
var
 OName             : string;
 pOfs              : TDefValue;
 pCnt              : TAsmValue;
 objParam          : string;
begin
   if Source[CurLevel].StateLevel > 0 then begin
      ErrorMsg('Can not create objects inside [.IF], [.FOR] and/or [.WHILE]!');
   end;
   if MacroName <> '' then begin
      ErrorMsg('Objects can not be created inside another object or macro!');
   end;
   if ProcName <> '' then begin
      ErrorMsg('Can not create objects inside procedures!');
   end;
   if SecName <> '' then begin
      ErrorMsg('Can not create objects inside sections!');
   end;
   if not IsErr then begin
      OName:= '';
      if objName <> '' then begin
         OName:= objName;
      end else begin
         OName:= GetName;
      end;
      if OName <> '' then begin
         pOfs:= ParamCnt;
         pCnt:= 0;
         objParam:= GetName();
         if objParam <> '' then begin
            AddParam(objParam);
            if not IsErr then begin
               SkipBlank;
               if GetSrcChar = '(' then begin
                  NextChar;
                  pCnt:= GrabParams(pOfs, true);
                  SkipBlank;
                  if GetSrcChar = ')' then begin
                     NextChar;
                  end else begin
                     ErrorMsg('[ ) ] expected!');
                  end;
               end else begin
                  ErrorMsg('[ ( ] expected!');
               end;
            end;
         end else begin
            ErrorMsg('Object name parameter expected!');
         end;
         if not IsErr then begin
            MacroName:= OName;
            MacroIsObject:= true;
            MacroStart:= Source[CurLevel].SrcOffset + Source[CurLevel].CurLine;
            MacroLength:= 0;
            MacroParamOffset:= pOfs;
            MacroParamCount:= pCnt;
         end;
      end else begin
         ErrorMsg('Object name expected!');
      end;
   end;
end;

function GrabParamValues(prmOfs: TDefValue): TAsmValue;
var
 prCnt             : TAsmValue;
 prStr             : string;
 inStr             : boolean;
 strChar           : char;
begin
   prCnt:= 0;
   prStr:= '';
   inStr:= false;
   strChar:= #39;
   SkipBlank;
   while true do begin
      if Source[CurLevel].Pos > Source[CurLevel].Len then begin
         if not inStr then begin
            AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
         end else begin
            ErrorMsg('[ ' + strChar + ' ] expected!');
         end;
         break;
      end;
      if GetSrcChar = ';' then begin
         if not inStr then begin
            AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
            break;
         end;
      end;
      if GetSrcChar = ',' then begin
         if not inStr then begin
            NextChar;
            SkipBlank;
            AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
            continue;
         end;
      end;
      if (GetSrcChar = '"') or (GetSrcChar = #39) then begin
         if not inStr then begin
            strChar:= GetSrcChar;
            inStr:= true;
         end else begin
            if GetSrcChar = strChar then begin
               inStr:= false;
            end;
         end;
      end;
      prStr:= prStr + GetSrcChar;
      NextChar;
   end;
   GrabParamValues:= prCnt;
end;

function GrabObjectParamValues(prmOfs: TDefValue): TAsmValue;
var
 prCnt             : TAsmValue;
 prStr             : string;
 inStr             : boolean;
 strChar           : char;
 BrackPos          : TDefValue;
begin
   prCnt:= 1;
   prStr:= '';
   inStr:= false;
   strChar:= #39;
   BrackPos:= 0;
   SkipBlank;
   while true do begin
      if Source[CurLevel].Pos > Source[CurLevel].Len then begin
         if not inStr then begin
            AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
         end else begin
            ErrorMsg('[ ' + strChar + ' ] expected!');
         end;
         break;
      end;
      if GetSrcChar = '(' then begin
         if not inStr then begin
            Inc(BrackPos);
         end;
      end;
      if GetSrcChar = ')' then begin
         if not inStr then begin
            if BrackPos > 0 then begin
               Dec(BrackPos);
            end else begin
               AddParamValue(Trim(prStr));
               Inc(prCnt);
               prStr:= '';
               break;
            end;
         end;
      end;
      if GetSrcChar = ';' then begin
         if not inStr then begin
	          AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
            break;
         end;
      end;
      if GetSrcChar = ',' then begin
         if not inStr then begin
            NextChar;
            SkipBlank;
            AddParamValue(Trim(prStr));
            Inc(prCnt);
            prStr:= '';
            continue;
         end;
      end;
      if (GetSrcChar = '"') or (GetSrcChar = #39) then begin
         if not inStr then begin
            strChar:= GetSrcChar;
            inStr:= true;
         end else begin
            if GetSrcChar = strChar then begin
               inStr:= false;
            end;
         end;
      end;
      prStr:= prStr + GetSrcChar;
      NextChar;
   end;
   GrabObjectParamValues:= prCnt;
end;



function GetMacroParam(n: string;
 po: TDefValue; pc: TAsmValue;
 pvo: TDefValue; pvc: TAsmValue): string;
var
 mPar, nn          : string;
 pFound            : boolean;
 i                 : TDefValue;
begin
   mPar:= '';
   pFound:= false;
   nn:= UpperCase(n);
   i:= 0;
   while i < pc do begin
      if nn = ParamList[po + i] then begin
         pFound:= true;
         if i < pvc then begin
            mPar:= ParamValList[pvo+i];
            break;
         end;
      end;
      Inc(i);
   end;
   If not pFound then begin
      mPar:= nn;
   end;
   GetMacroParam:= mPar;
end;

function InsertParams(s: string): string;
var
 ss                : string;
 pos, len          : integer;
 curName           : string;
 inStr             : boolean;
 strChar           : char;
 curChar           : char;
 asc               : char;
 macName           : string;
 pOfs              : TDefValue;
 pCnt              : TAsmValue;
 id                : TDefValue;
 pValOfs           : TDefValue;
 pValCnt           : TAsmValue;
begin
   ss:= '';
   pos:= 1;
   len:= Length(s);
   curName:= '';
   inStr:= false;
   strChar:= #39;
   macName:= Source[CurLevel].FileName;
   pOfs:= 0;
   pCnt:= 0;
   id:= GetIdentIndex(macName);
   if id <> NotFound then begin
      pOfs:= IdentList[id].ParamOffset;
      pCnt:= IdentList[id].ParamCnt;
   end;
   pValOfs:= Source[CurLevel].ParamValOfs;
   pValCnt:= Source[CurLevel].ParamValCnt;
   while true do begin
      if pos > len then begin
         if curName <> '' then begin
            ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
         end;
         break;
      end;
      curChar:= s[pos];
      asc:= curChar;
      if (curChar = '&') and (not inStr) then begin
         if curName <> '' then begin
            ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
            curName:= '';
         end;
         Inc(pos);
         continue;
      end;
      if (curChar = '{') and (not inStr) then begin
         Inc(pos);
         if curName <> '' then begin
            ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
            curName:= '';
         end;
         while true do begin
            if pos > len then begin
               ErrorMsg('[ { ] without [ } ]!');
               break;
            end;
            curChar:= s[pos];
            if curChar = '{' then begin
               ErrorMsg('[ } ] expected, but another [ { ] found!');
               break;
            end;
            if curChar = '}' then begin
               Inc(pos);
               break;
            end else begin
               ss:= ss + curChar;
            end;
            Inc(pos);
         end;
         continue;
      end;
      if (curChar = #39) or (curChar = '"') then begin
         if not inStr then begin
            if curName <> '' then begin
               ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
               curName:= '';
            end;
            strChar:= asc;
            inStr:= true;
         end else begin
            if strChar = asc then begin
               inStr:= false;
            end;
         end;
      end;
      if IsName(curChar) then begin
         if not inStr then begin
            if curChar <> '~' then
             curName:= curName + curChar
            else begin
               if curName <> '' then begin
                  ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
                  curName:= '';
               end;
               ss:= ss + curChar;
            end;
         end else begin
            ss:= ss + curChar;
         end;
      end else begin
         if curName <> '' then begin
            ss:= ss + GetMacroParam(curName, pOfs, pCnt, pValOfs, pValCnt);
            curName:= '';
         end;
         ss:= ss + curChar;
      end;
      Inc(pos);
   end;
   InsertParams:= ss;
end;



procedure BuildOutput;
var
 hdrIndex          : TDefValue;
begin
   if CurOutput <> Neg then begin
      PrepareSegments;
      hdrIndex:= OutputList[CurOutput].Header;
      if ErrorCount = 0 then PrepareHeader(hdrIndex);
      if ErrorCount = 0 then BuildHeader(hdrIndex);
      if ErrorCount = 0 then BuildSegments(hdrIndex);
      if ErrorCount = 0 then begin
         { FillRoundOutput(); }
         if (OutputList[CurOutput].Limit > 0) and
          (OutputLen > OutputList[CurOutput].Limit) then
          FatalError('Output Size is larger than the limit!');
         if ErrorCount = 0 then WriteOutput;
      end;
   end else begin
      FatalError('No output selected!');
   end;
end;

end.
