{ Segments
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _seg;

interface

uses
 _common;

type
 TSegment = record
    Name           : string;
    IdentName      : string;
    Offset         : TAsmValue;
    ORG            : TAsmValue;
    RealSize       : TAsmValue;
    SegBits        : TAsmValue;
    Limit          : TAsmValue;
    FillSize       : TAsmValue;
    RoundSize      : TAsmValue;
    WriteSEG       : boolean;
    FillPatternOfs : TDefValue;
    FillPatternLen : TAsmValue;
    FillPatternPos : TAsmValue;
 end;

var
 StackMem          : TAsmValue;

 Segment           : array of TSegment;
 SegCount          : TDefValue;
 CurSeg            : TDefValue;

 SegBytes          : array of byte;
 SegByteCount      : TAsmValue;

procedure InitSegments;
procedure AddSegment(segN, segIdent: string;
 flOfs: TDefValue; flLen: TAsmValue);

function GetSegIndex(n: string): TDefValue;

function GetSegOffset(useORG: boolean): TAsmValue;
function GetSeg: TDefValue;
procedure AddSegSize(siz: TAsmValue);

procedure AddSegByte(b: byte);

procedure PrepareSegments;
procedure BuildSegments(indx: TDefValue);

implementation

uses
 SysUtils, _error, _emit, _func, _conv, _scanner,
 _parser, _ident, _output, _sec, _codegen, _mem, _header;


procedure InitSegments;
begin
   StackMem:= 0;
   SegCount:= 0;
   CurSeg:= Neg;
   SetLength(SegBytes, 1);
   SegByteCount:= 0;
end;

procedure AddSegment(segN, segIdent: string;
 flOfs: TDefValue; flLen: TAsmValue);
var
 i                 : TDefValue;
begin
   if SegCount > 0 then begin
      i:= 0;
      while i < SegCount do begin
         if Segment[i].IdentName = segIdent then begin
            ErrorMsg('Segment already exists [' + segIdent + ']!');
            break;
         end;
         Inc(i);
      end;
   end;
   if not IsErr then begin
      SetLength(Segment, SegCount + 1);
      Segment[SegCount].Name:= segN;
      Segment[SegCount].IdentName:= segIdent;
      Segment[SegCount].Offset:= 0;
      Segment[SegCount].ORG:= 0;
      Segment[SegCount].RealSize:= 0;
      Segment[SegCount].SegBits:= 0;
      Segment[SegCount].Limit:= 0;
      Segment[SegCount].FillSize:= 0;
      Segment[SegCount].RoundSize:= 0;
      Segment[SegCount].WriteSEG:= true;
      Segment[SegCount].FillPatternOfs:= 0;
      Segment[SegCount].FillPatternLen:= 0;
      Segment[SegCount].FillPatternPos:= 0;
      Inc(SegCount);
      if GetSeg = Neg then CurSeg:= 0;
   end;
end;



function GetSegIndex(n: string): TDefValue;
var
 i, id             : TDefValue;
begin
   id:= NotFound;
   if SegCount > 0 then begin
      i:= 0;
      while i < SegCount do begin
         if Segment[i].IdentName = n then begin
            id:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetSegIndex:= id;
end;



function GetSegOffset(useORG: boolean): TAsmValue;
var
 rOfs              : TAsmValue;
begin
   rOfs:= 0;
   if not useORG then begin
      if SecName = '' then begin
         if CurSeg <> Neg then begin
            rOfs:= Segment[CurSeg].Offset;
         end;
      end else begin
         rOfs:= SecPos;
      end;
   end else begin
      if SecName = '' then begin
         if CurSeg <> Neg then begin
            rOfs:= Segment[CurSeg].Offset + Segment[CurSeg].ORG;
         end;
      end else begin
         rOfs:= SecPos + SecORG;
      end;
   end;
   GetSegOffset:= rOfs;
end;

function GetSeg: TDefValue;
var
 gSeg              : TDefValue;
begin
   gSeg:= CurSeg;
   if SecName <> '' then gSeg:= SecSegment;
   GetSeg:= gSeg;
end;

procedure AddSegSize(siz: TAsmValue);
var
 cSeg              : TDefValue;
begin
   cSeg:= GetSeg;
   if cSeg <> Neg then begin
      if SecName <> '' then SecPos:= SecPos + siz;
      Segment[cSeg].Offset:= Segment[cSeg].Offset + siz;
      Segment[cSeg].RealSize:= Segment[cSeg].RealSize + siz;
   end;
end;



procedure AddSegByte(b: byte);
begin
   SetLength(SegBytes, SegByteCount + 1);
   SegBytes[SegByteCount]:= b;
   Inc(SegByteCount);
end;



procedure PrepareSegments;
var
 seg               : TDefValue;
 CurSize           : TAsmValue;
 FillCount         : TAsmValue;
 RoundCount        : TAsmValue;
 ptnOfs            : TDefValue;
 ptnLen            : TAsmValue;
 ptnPos            : TAsmValue;
 EndSeg, iFnd      : boolean;
 SegPos, ofs, nVal : TAsmValue;
 i, ii, si, f, r   : TDefValue;
 id                : TDefValue;
 ident             : string;
 index             : TDefValue;
 val1{, val2    }  : TAsmValue;
 opr               : byte;
 vt1{, vt2}        : byte;
 siz               : byte;
 b                 : byte;
 iMod              : byte;
 {strOfs, strLen    : TAsmValue;}
begin
   if SegCount > 0 then begin
      EmitLn('------ START MAKING SEGMENT DATA [' + MakeTime(true) + '] ------');
      seg:= 0;
      while seg < SegCount do begin
         FillCount:= 0;
         RoundCount:= 0;
         if Segment[seg].FillSize > 0 then begin
            if Segment[seg].Offset < Segment[seg].FillSize then begin
               FillCount:= Segment[seg].FillSize - Segment[seg].Offset;
               Segment[seg].Offset:= Segment[seg].Offset + FillCount;
            end;
         end;
         if Segment[seg].RoundSize > 0 then begin
            CurSize:= Segment[seg].Offset;
            RoundCount:= CurSize div Segment[seg].RoundSize;
            RoundCount:= CurSize - (RoundCount * Segment[seg].RoundSize);
            if (RoundCount > 0) and (RoundCount < Segment[seg].RoundSize) then
             begin
               RoundCount:= Segment[seg].RoundSize - RoundCount;
               Segment[seg].Offset:= Segment[seg].Offset + RoundCount;
            end;
         end;

         EndSeg:= false;
         SegPos:= 0;
         if CodeListItems > 0 then begin
            iMod:= IdentMode;
            IdentMode:= imLOCAL;
            i:= 0;
            while i < CodeListItems do begin
               IsErr:= false;
	             if EndSeg then break;
               if CodeList[i].SegIndex = seg then begin
                  id:= CodeList[i].ID;
                  ident:= CodeList[i].IdentName;
                  index:= CodeList[i].Index;
                  val1:= CodeList[i].Value1;
                  {val2:= CodeList[i].Value2;}
                  opr:= CodeList[i].Opr;
                  {strOfs:= CodeList[i].StringOfs;
                  strLen:= CodeList[i].StringLen;}
                  vt1:= CodeList[i].Type1;
                  {vt2:= CodeList[i].Type2;}
                  iFnd:= false;
                  if (id = codeLOCALMODE) and (not iFnd) then begin
                     iFnd:= true;
                     IdentMode:= imLOCAL;
                  end;
                  if (id = codeGLOBALMODE) and (not iFnd) then begin
                     iFnd:= true;
                     IdentMode:= imGLOBAL;
                  end;
                  if (id = codeVALUE) and (not iFnd) then begin
                     iFnd:= true;
                     siz:= 1;
                     if vt1 = tWORD then siz:= 2;
                     if vt1 = tDWORD then siz:= 4;
                     if vt1 = tQWORD then siz:= 8;
                     if vt1 = tFARPTR then siz:= 6;
                     if vt1 = tTENBYTE then siz:= 10;
                     ValueToBytes(val1);
                     SetLength(SegBytes, SegByteCount + siz + 1);
                     ii:= 0;
                     while ii < siz do begin
                        SegBytes[SegByteCount+ii]:= byt[ii+1];
                        Inc(ii);
                     end;
                     SegByteCount:= SegByteCount + siz;
                     SegPos:= SegPos + siz;
                  end;
                  if (id = codeREL) and (not iFnd) then begin
                     iFnd:= true;
                     { ofs:= 0; }
                     index:= GetIdentIndex(ident);
                     if index <> NotFound then begin
                        siz:= 1;
                        if vt1 = tWORD then siz:= 2;
                        if vt1 = tDWORD then siz:= 4;
                        if vt1 = tQWORD then siz:= 8;
                        if vt1 = tFARPTR then siz:= 6;
                        if vt1 = tTENBYTE then siz:= 10;
                        ofs:= IdentList[index].Offset;
                        if opr <> oprNONE then
                         ofs:= DoCalc(IdentList[index].Offset, opr, val1);
                        if SegPos > ofs then begin
                           nVal:= $ffffffff - 3;
                           if vt1 = tBYTE then nVal:= 255;
                           if vt1 = tWORD then nVal:= 65534;
                           ofs:= nVal - (SegPos - ofs);
                           if (vt1 = tBYTE) and (ofs < 128) then FatalError('8-bit Relative Address out of range [' + ident + ']!');
                           if (vt1 = tWORD) and (ofs < 32768) then FatalError('16-bit Relative Address out of range [' + ident + ']!');
                           if (vt1 = tDWORD) and (ofs < 2147483648) then FatalError('32-bit Relative Address out of range [' + ident + ']!');
                        end else begin
                           ofs:= ofs - (SegPos + siz);
                           if (vt1 = tBYTE) and (ofs > 127) then FatalError('8-bit Relative Address out of range [' + ident + ']!');
                           if (vt1 = tWORD) and (ofs > 32767) then FatalError('16-bit Relative Address out of range [' + ident + ']!');
                           if (vt1 = tDWORD) and (ofs > 2147483647) then FatalError('32-bit Relative Address out of range [' + ident + ']!');
                        end;
                        if siz > 4 then
                         FatalError('Only 8-, 16- or 32-bit Relative addresses allowed!');
                        if not IsErr then begin
                           ValueToBytes(ofs);
                           SetLength(SegBytes, SegByteCount + siz + 1);
                           ii:= 0;
                           while ii < siz do begin
                              SegBytes[SegByteCount+ii]:= byt[ii+1];
                              Inc(ii);
                           end;
                           SegByteCount:= SegByteCount + siz;
                           SegPos:= SegPos + siz;
                        end;
                     end else begin
                        FatalError('Identifier not found [' + ident + ']!');
                     end;
                  end;
                  if (id = codeRELID) and (not iFnd) then begin
                     iFnd:= true;
                     {ofs:= 0;}
		                 siz:= 1;
                     if vt1 = tWORD then siz:= 2;
                     if vt1 = tDWORD then siz:= 4;
                     if vt1 = tQWORD then siz:= 8;
                     if vt1 = tFARPTR then siz:= 6;
                     if vt1 = tTENBYTE then siz:= 10;
                     if siz > 4 then
                      FatalError('Only 8-, 16- or 32-bit Relative addresses allowed!');
                     if not IsErr then begin
                        ofs:= IdentList[index].Offset;
                        if opr <> oprNONE then
                         ofs:= DoCalc(IdentList[index].Offset, opr, val1);
                        if SegPos > ofs then begin
                           nVal:= $ffffffff - 3;
                           if vt1 = tBYTE then nVal:= 255;
                           if vt1 = tWORD then nVal:= 65534;
                           ofs:= nVal - (SegPos - ofs);
                           if (vt1 = tBYTE) and (ofs < 128) then FatalError('8-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                           if (vt1 = tWORD) and (ofs < 32768) then FatalError('16-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                           if (vt1 = tDWORD) and (ofs < 2147483648) then FatalError('32-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                        end else begin
                           ofs:= ofs - (SegPos + siz);
                           if (vt1 = tBYTE) and (ofs > 127) then FatalError('8-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                           if (vt1 = tWORD) and (ofs > 32767) then FatalError('16-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                           if (vt1 = tDWORD) and (ofs > 2147483647) then FatalError('32-bit Relative Address out of range [' + IdentList[index].Name + ']!');
                        end;
                        if not IsErr then begin
                           ValueToBytes(ofs);
                           SetLength(SegBytes, SegByteCount + siz);
                           ii:= 0;
                           while ii < siz do begin
                              SegBytes[SegByteCount+ii]:= byt[ii+1];
                              Inc(ii);
                           end;
                           SegByteCount:= SegByteCount + siz;
                           SegPos:= SegPos + siz;
                        end;
                     end;
                  end;
                  if (id = codeMEM) and (not iFnd) then begin
                     iFnd:= true;
                     { ofs:= 0; }
                     index:= GetIdentIndex(ident);
                     if index <> NotFound then begin
                        siz:= 1;
                        if vt1 = tWORD then siz:= 2;
                        if vt1 = tDWORD then siz:= 4;
                        if vt1 = tQWORD then siz:= 8;
                        if vt1 = tFARPTR then siz:= 6;
                        if vt1 = tTENBYTE then siz:= 10;
                        ofs:= IdentList[index].Offset;
                        if opr <> oprNONE then
                         ofs:= DoCalc(IdentList[index].Offset, opr, val1);
                        ValueToBytes(ofs);
                        SetLength(SegBytes, SegByteCount + siz);
                        ii:= 0;
                        while ii < siz do begin
                           SegBytes[SegByteCount+ii]:= byt[ii+1];
                           Inc(ii);
                        end;
                        SegByteCount:= SegByteCount + siz;
                        SegPos:= SegPos + siz;
                     end else begin
                        FatalError('Identifier not found [' + ident + ']!');
                     end;
                  end;
                  if (id = codeMEMID) and (not iFnd) then begin
		                 iFnd:= true;
		                 siz:= 1;
                     if vt1 = tWORD then siz:= 2;
                     if vt1 = tDWORD then siz:= 4;
                     if vt1 = tQWORD then siz:= 8;
                     if vt1 = tFARPTR then siz:= 6;
                     if vt1 = tTENBYTE then siz:= 10;
                     ofs:= IdentList[index].Offset;
                     if opr <> oprNONE then
                      ofs:= DoCalc(IdentList[index].Offset, opr, val1);
                     ValueToBytes(ofs);
                     SetLength(SegBytes, SegByteCount + siz);
                     ii:= 0;
                     while ii < siz do begin
                        SegBytes[SegByteCount+ii]:= byt[ii+1];
                        Inc(ii);
                     end;
                     SegByteCount:= SegByteCount + siz;
                     SegPos:= SegPos + siz;
                  end;
                  if (id = codeEQUOFFSET) and (not iFnd) then begin
                     iFnd:= true;
                     FatalError('Identifier not found [' + ident + ']!');
                  end;
                  if (id = codeEQUSIZE) and (not iFnd) then begin
                     iFnd:= true;
                     FatalError('Identifier not found [' + ident + ']!');
                  end;
                  if (id = codeGETSEGOFFSET) and (not iFnd) then begin
                     iFnd:= true;
                     ofs:= 0;
                     si:= 0;
                     while si < index do begin
                        ofs:= ofs + Segment[si].Offset;
                        Inc(si);
                     end;
		                 siz:= 1;
                     if vt1 = tWORD then siz:= 2;
                     if vt1 = tDWORD then siz:= 4;
                     if vt1 = tQWORD then siz:= 8;
                     if vt1 = tFARPTR then siz:= 6;
                     if vt1 = tTENBYTE then siz:= 10;
                     if opr <> oprNONE then
                      ofs:= DoCalc(ofs, opr, val1);
                     ValueToBytes(ofs);
                     SetLength(SegBytes, SegByteCount + siz);
                     ii:= 0;
                     while ii < siz do begin
                        SegBytes[SegByteCount+ii]:= byt[ii+1];
                        Inc(ii);
                     end;
                     SegByteCount:= SegByteCount + siz;
                     SegPos:= SegPos + siz;
                  end;
                  if (id = codeGETSEGSIZE) and (not iFnd) then begin
                     iFnd:= true;
                     ofs:= Segment[index].Offset;
                     siz:= 1;
                     if vt1 = tWORD then siz:= 2;
                     if vt1 = tDWORD then siz:= 4;
                     if vt1 = tQWORD then siz:= 8;
                     if vt1 = tFARPTR then siz:= 6;
                     if vt1 = tTENBYTE then siz:= 10;
                     if opr <> oprNONE then
                      ofs:= DoCalc(ofs, opr, val1);
                     ValueToBytes(ofs);
                     SetLength(SegBytes, SegByteCount + siz);
                     ii:= 0;
                     while ii < siz do begin
                        SegBytes[SegByteCount+ii]:= byt[ii+1];
                        Inc(ii);
                     end;
                     SegByteCount:= SegByteCount + siz;
                     SegPos:= SegPos + siz;
                  end;
                  if (id = codeEND) and (not iFnd) then begin
                     iFnd:= true;
                     EndSeg:= true;
                  end;
                  if (id = codeRESETSEG) and (not iFnd) then begin
                     iFnd:= true;
		                 { *** Reset Segment *** }
                  end;
		              if not iFnd then begin
                     FatalError('Invalid Code ID [' + IntToStr(id) + ']!');
                  end;
               end;
               Inc(i);
            end;
            IdentMode:= iMod;
         end;

         ptnOfs:= Segment[seg].FillPatternOfs;
         ptnLen:= Segment[seg].FillPatternLen;
         ptnPos:= Segment[seg].FillPatternPos;
         if Segment[seg].WriteSEG then begin
            f:= 0;
            while f < FillCount do begin
	             b:= 0;
               if ptnLen > 0 then begin
                  b:= MemData[ptnOfs + ptnPos];
                  Inc(ptnPos);
                  if ptnPos > (ptnLen-1) then ptnPos:= 0;
               end;
               AddSegByte(b);
               Inc(f);
            end;
            r:= 0;
            while r < RoundCount do begin
               b:= 0;
               if ptnLen > 0 then begin
                  b:= MemData[ptnOfs + ptnPos];
                  Inc(ptnPos);
                  if ptnPos > (ptnLen-1) then ptnPos:= 0;
               end;
               AddSegByte(b);
               Inc(r);
            end;
         end;
         if Segment[seg].Limit > 0 then begin
            if Segment[seg].Offset > Segment[seg].Limit then begin
               FatalError('Segment is larger than the limit [' +
                Segment[seg].IdentName + ']!');
            end;
         end;

         EmitLn(Segment[seg].Name + ', Size: ' +
          IntToStr(Segment[seg].RealSize) + ', Total Size: ' +
          IntToStr(Segment[seg].Offset));

         Inc(seg);
      end;
      EmitLn('------ END MAKING SEGMENT DATA [' + MakeTime(true) + '] ------');
   end;
end;

procedure BuildSegments(indx: TDefValue);
var
 i                 : TDefValue;
 stOfs             : TAsmValue;
 b                 : byte;
begin
   EmitLn('------ START GENERATING OUTPUT FILE [' + MakeTime(true) + '] ------');
   stOfs:= HeaderList[indx].HeaderSize;
   if SegByteCount > 0 then begin
      SetLength(OutputMem, stOfs+SegByteCount);
      i:= 0;
      while i < SegByteCount do begin
         b:= SegBytes[i];
         OutputMem[stOfs+i]:= b;
         Inc(i);
      end;
   end;
   FillRoundOutput(stOfs+SegByteCount, WrittenBytes);
   EmitLn('------ END GENERATING OUTPUT FILE [' + MakeTime(true) + '] ------');
end;

end.
 
